from __future__ import annotations

import json
import os
from tempfile import mkstemp
from typing import TYPE_CHECKING

import click
from typing_extensions import TypedDict, Unpack

from pyinfra import local, logger
from pyinfra.api import QuoteString, StringCommand
from pyinfra.api.exceptions import ConnectError, InventoryError, PyinfraError
from pyinfra.api.util import get_file_io
from pyinfra.progress import progress_spinner

from .base import BaseConnector, DataMeta
from .local import LocalConnector
from .util import CommandOutput, extract_control_arguments, make_unix_command_for_host

if TYPE_CHECKING:
    from pyinfra.api.arguments import ConnectorArguments
    from pyinfra.api.host import Host
    from pyinfra.api.state import State


class ConnectorData(TypedDict):
    docker_identifier: str


connector_data_meta: dict[str, DataMeta] = {
    "docker_identifier": DataMeta("ID of container or image to start from"),
}


def _find_start_docker_container(container_id) -> tuple[str, bool]:
    docker_info = local.shell("docker container inspect {0}".format(container_id))
    assert isinstance(docker_info, str)
    docker_info = json.loads(docker_info)[0]
    if docker_info["State"]["Running"] is False:
        logger.info("Starting stopped container: {0}".format(container_id))
        local.shell("docker container start {0}".format(container_id))
        return container_id, False
    return container_id, True


def _start_docker_image(image_name):
    try:
        return local.shell(
            "docker run -d {0} tail -f /dev/null".format(image_name),
            splitlines=True,
        )[
            -1
        ]  # last line is the container ID
    except PyinfraError as e:
        raise ConnectError(e.args[0])


class DockerConnector(BaseConnector):
    """
    The docker connector allows you to build Docker images or modify running
    Docker containers. You can pass either an image name or existing container ID:

    + Image - will create a new container from the image, execute operations \
        against it, save into a new Docker image and remove the container
    + Existing container ID - will execute operations against the running \
        container, leaving it running

    .. code:: shell

        # A Docker base image must be provided
        pyinfra @docker/alpine:3.8 ...

        # pyinfra can run on multiple Docker images in parallel
        pyinfra @docker/alpine:3.8,@docker/ubuntu:bionic ...

        # Execute against a running container
        pyinfra @docker/2beb8c15a1b1 ...
    """

    handles_execution = True

    data_cls = ConnectorData
    data_meta = connector_data_meta
    data: ConnectorData

    local: LocalConnector

    container_id: str
    no_stop: bool = False

    def __init__(self, state: "State", host: "Host"):
        super().__init__(state, host)
        self.local = LocalConnector(state, host)

    @staticmethod
    def make_names_data(name=None):
        if not name:
            raise InventoryError("No docker base ID provided!")

        yield (
            "@docker/{0}".format(name),
            {"docker_identifier": name},
            ["@docker"],
        )

    def connect(self) -> None:
        self.local.connect()

        docker_identifier = self.data["docker_identifier"]
        with progress_spinner({"prepare docker container"}):
            try:
                self.container_id, was_running = _find_start_docker_container(docker_identifier)
                if was_running:
                    self.no_stop = True
            except PyinfraError:
                self.container_id = _start_docker_image(docker_identifier)

    def disconnect(self):
        container_id = self.container_id

        if self.no_stop:
            logger.info(
                "{0}docker build complete, container left running: {1}".format(
                    self.host.print_prefix,
                    click.style(container_id, bold=True),
                ),
            )
            return

        with progress_spinner({"docker commit"}):
            image_id = local.shell("docker commit {0}".format(container_id), splitlines=True)[-1][
                7:19
            ]  # last line is the image ID, get sha256:[XXXXXXXXXX]...

        with progress_spinner({"docker rm"}):
            local.shell(
                "docker rm -f {0}".format(container_id),
            )

        logger.info(
            "{0}docker build complete, image ID: {1}".format(
                self.host.print_prefix,
                click.style(image_id, bold=True),
            ),
        )

    def run_shell_command(
        self,
        command: StringCommand,
        print_output: bool = False,
        print_input: bool = False,
        **arguments: Unpack["ConnectorArguments"],
    ) -> tuple[bool, CommandOutput]:
        local_arguments = extract_control_arguments(arguments)

        container_id = self.container_id

        command = make_unix_command_for_host(self.state, self.host, command, **arguments)
        command = StringCommand(QuoteString(command))

        docker_flags = "-it" if local_arguments.get("_get_pty") else "-i"
        docker_command = StringCommand(
            "docker",
            "exec",
            docker_flags,
            container_id,
            "sh",
            "-c",
            command,
        )

        return self.local.run_shell_command(
            docker_command,
            print_output=print_output,
            print_input=print_input,
            **local_arguments,
        )

    def put_file(
        self,
        filename_or_io,
        remote_filename,
        remote_temp_filename=None,  # ignored
        print_output=False,
        print_input=False,
        **kwargs,  # ignored (sudo/etc)
    ) -> bool:
        """
        Upload a file/IO object to the target Docker container by copying it to a
        temporary location and then uploading it into the container using ``docker cp``.
        """

        fd, temp_filename = mkstemp()

        try:
            # Load our file or IO object and write it to the temporary file
            with get_file_io(filename_or_io) as file_io:
                with open(temp_filename, "wb") as temp_f:
                    data = file_io.read()

                    if isinstance(data, str):
                        data = data.encode()

                    temp_f.write(data)

            docker_command = StringCommand(
                "docker",
                "cp",
                temp_filename,
                f"{self.container_id}:{remote_filename}",
            )

            status, output = self.local.run_shell_command(
                docker_command,
                print_output=print_output,
                print_input=print_input,
            )
        finally:
            os.close(fd)
            os.remove(temp_filename)

        if not status:
            raise IOError(output.stderr)

        if print_output:
            click.echo(
                "{0}file uploaded to container: {1}".format(
                    self.host.print_prefix,
                    remote_filename,
                ),
                err=True,
            )

        return status

    def get_file(
        self,
        remote_filename,
        filename_or_io,
        remote_temp_filename=None,  # ignored
        print_output=False,
        print_input=False,
        **kwargs,  # ignored (sudo/etc)
    ) -> bool:
        """
        Download a file from the target Docker container by copying it to a temporary
        location and then reading that into our final file/IO object.
        """

        fd, temp_filename = mkstemp()

        try:
            docker_command = StringCommand(
                "docker",
                "cp",
                f"{self.container_id}:{remote_filename}",
                temp_filename,
            )

            status, output = self.local.run_shell_command(
                docker_command,
                print_output=print_output,
                print_input=print_input,
            )

            # Load the temporary file and write it to our file or IO object
            with open(temp_filename, "rb") as temp_f:
                with get_file_io(filename_or_io, "wb") as file_io:
                    data = temp_f.read()
                    file_io.write(data)
        finally:
            os.close(fd)
            os.remove(temp_filename)

        if not status:
            raise IOError(output.stderr)

        if print_output:
            click.echo(
                "{0}file downloaded from container: {1}".format(
                    self.host.print_prefix,
                    remote_filename,
                ),
                err=True,
            )

        return status



from __future__ import annotations

import re
from typing import Dict, Iterable

from pyinfra.api import FactBase, QuoteString, StringCommand

# Valid unit names consist of a "name prefix" and a dot and a suffix specifying the unit type.
# The "unit prefix" must consist of one or more valid characters
# (ASCII letters, digits, ":", "-", "_", ".", and "\").
# The total length of the unit name including the suffix must not exceed 256 characters.
# The type suffix must be one of
# ".service", ".socket", ".device", ".mount", ".automount",
# ".swap", ".target", ".path", ".timer", ".slice", or ".scope".
# Units names can be parameterized by a single argument called the "instance name".
# A template unit must have a single "@" at the end of the name (right before the type suffix).
# The name of the full unit is formed by inserting the instance name
# between "@" and the unit type suffix.
SYSTEMD_UNIT_NAME_REGEX = (
    r"[a-zA-Z0-9\:\-\_\.\\\@]+\."
    r"(?:service|socket|device|mount|automount|swap|target|path|timer|slice|scope)"
)


def _make_systemctl_cmd(user_mode=False, machine=None, user_name=None):
    # base command for normal and user mode
    systemctl_cmd = ["systemctl --user"] if user_mode else ["systemctl"]

    # add user and machine flag if given in args
    if machine is not None:
        if user_name is not None:
            systemctl_cmd.append("--machine={1}@{0}".format(machine, user_name))
        else:
            systemctl_cmd.append("--machine={0}".format(machine))
    elif user_name is not None:
        # If only the user is given, assume that the connection should be made to the local machine
        systemctl_cmd.append("--machine={0}@.host".format(user_name))

    return StringCommand(*systemctl_cmd)


class SystemdStatus(FactBase[Dict[str, bool]]):
    """
    Returns a dictionary map of systemd units to booleans indicating whether they are active.

    + user_mode: whether to use user mode
    + machine: machine name

    .. code:: python

        {
            "ssh.service": True,
            "containerd.service": True,
            "apt-daily.timer": False,
        }
    """

    def requires_command(self, *args, **kwargs) -> str:
        return "systemctl"

    default = dict

    state_key = "SubState"
    state_values = ["running", "waiting", "exited", "listening"]

    def command(
        self,
        user_mode: bool = False,
        machine: str | None = None,
        user_name: str | None = None,
        services: str | list[str] | None = None,
    ) -> StringCommand:
        fact_cmd = _make_systemctl_cmd(
            user_mode=user_mode,
            machine=machine,
            user_name=user_name,
        )

        if services is None:
            service_strs = [QuoteString("*")]
        elif isinstance(services, str):
            service_strs = [QuoteString(services)]
        elif isinstance(services, Iterable):
            service_strs = [QuoteString(s) for s in services]

        return StringCommand(
            fact_cmd,
            "show",
            "--all",
            "--property",
            "Id",
            "--property",
            self.state_key,
            *service_strs,
        )

    def process(self, output) -> Dict[str, bool]:
        services: Dict[str, bool] = {}

        current_unit = None
        for line in output:
            line = line.strip()

            try:
                key, value = line.split("=", 1)
            except ValueError:
                current_unit = None  # reset current_unit just in case
                continue

            if key == "Id" and re.match(SYSTEMD_UNIT_NAME_REGEX, value):
                current_unit = value
                continue

            if key == self.state_key and current_unit:
                services[current_unit] = value in self.state_values

        return services


class SystemdEnabled(SystemdStatus):
    """
    Returns a dictionary map of systemd units to booleans indicating whether they are enabled.

    .. code:: python

        {
            "ssh.service": True,
            "containerd.service": True,
            "apt-daily.timer": False,
        }
    """

    state_key = "UnitFileState"
    state_values = ["enabled", "static"]



"""
Manage ZFS filesystems.
"""

from pyinfra import host
from pyinfra.api import operation
from pyinfra.facts.zfs import Datasets, Snapshots


@operation()
def dataset(
    dataset_name,
    present=True,
    recursive=False,
    sparse=None,
    volume_size=None,
    properties={},
    **extra_props,
):
    """
    Create, destroy or set properties on a  ZFS dataset (e.g. filesystem,
    volume, snapshot).

    + dataset_name: name of the filesystem to operate on
    + present: whether the named filesystem should exist
    + recursive: whether to create parent datasets, or destroy child datasets
    + sparse: for volumes, whether to create a sparse volume with no allocation
    + volume_size: the size of the volume
    + properties: the ZFS properties that should be set on the dataset.
    + **extra_props: additional props; merged with `properties` for convenience

    **Examples:**

    .. code:: python

        zfs.dataset(
            "tank/srv",
            mountpoint="/srv",
            compression="lz4",
            properties={"com.sun:auto_snapshot": "true"}
        )
        zfs.dataset("tank/vm-disks/db_srv_04", volume_size="32G") # creates a volume
        zfs.dataset("tank/home@old_version", present=False)

    """

    noop_msg = "{0} is already {1}".format(dataset_name, "present" if present else "absent")

    properties.update(extra_props)

    datasets = host.get_fact(Datasets)

    existing_dataset = datasets.get(dataset_name)

    if present and not existing_dataset:
        args = ["-o {0}={1}".format(prop, value) for prop, value in properties.items()]
        if recursive:
            args.append("-p")
        if sparse:
            args.append("-s")
        if volume_size:
            args.append("-V {0}".format(volume_size))

        args.sort()  # dicts are unordered, so make sure the test results are deterministic

        yield "zfs create {0} {1}".format(" ".join(args), dataset_name)

    elif present and existing_dataset:
        prop_args = [
            "{0}={1}".format(prop, value)
            for prop, value in properties.items() - existing_dataset.items()
        ]
        prop_args.sort()
        if prop_args:
            yield "zfs set {0} {1}".format(" ".join(prop_args), dataset_name)
        else:
            host.noop(noop_msg)

    elif existing_dataset and not present:
        recursive_arg = "-r" if recursive else ""
        yield "zfs destroy {0} {1}".format(recursive_arg, dataset_name)

    else:
        host.noop(noop_msg)


@operation()
def snapshot(snapshot_name, present=True, recursive=False, properties={}, **extra_props):
    """
    Create or destroy a ZFS snapshot, or modify its properties.

    + dataset_name: name of the filesystem to operate on
    + present: whether the named filesystem should exist
    + recursive: whether to snapshot child datasets
    + properties: the ZFS properties that should be set on the snapshot.
    + **extra_props: additional props; merged with `properties` for convenience

    **Examples:**

    .. code:: python

        zfs.snapshot("tank/home@weekly_backup")

    """
    properties.update(extra_props)
    snapshots = host.get_fact(Snapshots)

    if snapshot_name in snapshots or not present:
        yield from dataset._inner(snapshot_name, present=present, properties=properties)

    else:
        args = ["-o {0}={1}".format(prop, value) for prop, value in properties.items()]
        if recursive:
            args.append("-r")
        yield "zfs snap {0} {1}".format(" ".join(args), snapshot_name)


@operation()
def volume(
    volume_name, size, sparse=False, present=True, recursive=False, properties={}, **extra_props
):
    """
    Create or destroy a ZFS volume, or modify its properties.

    + volume_name: name of the volume to operate on
    + size: the size of the volume
    + sparse: create a sparse volume
    + present: whether the named volume should exist
    + recursive: whether to create parent datasets or destroy child datasets
    + properties: the ZFS properties that should be set on the snapshot.
    + **extra_props: additional props; merged with `properties` for convenience

    **Examples:**

    .. code:: python

        zfs.volume("tank/vm-disks/db_srv_04", "32G")

    """
    properties.update(extra_props)
    yield from dataset._inner(
        volume_name,
        volume_size=size,
        present=present,
        sparse=sparse,
        recursive=recursive,
        properties=properties,
    )


@operation()
def filesystem(fs_name, present=True, recursive=False, properties={}, **extra_props):
    """
    Create or destroy a ZFS filesystem, or modify its properties.

    + fs_name: name of the volume to operate on
    + present: whether the named volume should exist
    + recursive: whether to create parent datasets or destroy child datasets
    + properties: the ZFS properties that should be set on the snapshot.
    + **extra_props: additional props; merged with `properties` for convenience

    **Examples:**

    .. code:: python

        zfs.filesystem("tank/vm-disks/db_srv_04", "32G")

    """
    properties.update(extra_props)
    yield from dataset._inner(
        fs_name,
        present=present,
        recursive=recursive,
        properties=properties,
    )



"""
Manage npm (aka node aka Node.js) packages.
"""

from __future__ import annotations

from pyinfra import host
from pyinfra.api import operation
from pyinfra.facts.npm import NpmPackages

from .util.packaging import ensure_packages


@operation()
def packages(
    packages: str | list[str] | None = None,
    present=True,
    latest=False,
    directory: str | None = None,
):
    """
    Install/remove/update npm packages.

    + packages: list of packages to ensure
    + present: whether the packages should be present
    + latest: whether to upgrade packages without a specified version
    + directory: directory to manage packages for, defaults to global

    Versions:
        Package versions can be pinned like npm: ``<pkg>@<version>``.
    """

    current_packages = host.get_fact(NpmPackages, directory=directory)

    install_command = (
        "npm install -g" if directory is None else "cd {0} && npm install".format(directory)
    )

    uninstall_command = (
        "npm uninstall -g" if directory is None else "cd {0} && npm uninstall".format(directory)
    )

    upgrade_command = (
        "npm update -g" if directory is None else "cd {0} && npm update".format(directory)
    )

    yield from ensure_packages(
        host,
        packages,
        current_packages,
        present,
        install_command=install_command,
        uninstall_command=uninstall_command,
        upgrade_command=upgrade_command,
        version_join="@",
        latest=latest,
    )



"""
Operations are the core of pyinfra. The ``@operation`` wrapper intercepts calls
to the function and instead diff against the remote server, outputting commands
to the deploy state. This is then run later by pyinfra's ``__main__`` or the
:doc:`./pyinfra.api.operations` module.
"""

from __future__ import annotations

from functools import wraps
from inspect import signature
from io import StringIO
from types import FunctionType
from typing import TYPE_CHECKING, Any, Callable, Generator, Iterator, Optional, cast

from typing_extensions import ParamSpec

import pyinfra
from pyinfra import context, logger
from pyinfra.context import ctx_host, ctx_state

from .arguments import EXECUTION_KWARG_KEYS, AllArguments, pop_global_arguments
from .arguments_typed import PyinfraOperation
from .command import PyinfraCommand, StringCommand
from .exceptions import OperationValueError, PyinfraError
from .host import Host
from .operations import run_host_op
from .state import State, StateOperationHostData, StateOperationMeta
from .util import (
    get_call_location,
    get_file_sha1,
    get_operation_order_from_stack,
    log_operation_start,
    make_hash,
)

op_meta_default = object()

if TYPE_CHECKING:
    from pyinfra.connectors.util import CommandOutput


class OperationMeta:
    _hash: str

    _combined_output: Optional["CommandOutput"] = None
    _commands: Optional[list[Any]] = None
    _maybe_is_change: Optional[bool] = None
    _success: Optional[bool] = None

    def __init__(self, hash, is_change: Optional[bool]):
        self._hash = hash
        self._maybe_is_change = is_change

    def __repr__(self) -> str:
        """
        Return Operation object as a string.
        """

        if self._commands is not None:
            return (
                "OperationMeta(executed=True, "
                f"success={self.did_succeed()}, hash={self._hash}, commands={len(self._commands)})"
            )
        return (
            "OperationMeta(executed=False, "
            f"maybeChange={self._maybe_is_change}, hash={self._hash})"
        )

    # Completion & status checks
    def set_complete(
        self,
        success: bool,
        commands: list[Any],
        combined_output: "CommandOutput",
    ) -> None:
        if self.is_complete():
            raise RuntimeError("Cannot complete an already complete operation")
        self._success = success
        self._commands = commands
        self._combined_output = combined_output

    def is_complete(self) -> bool:
        return self._success is not None

    def _raise_if_not_complete(self) -> None:
        if not self.is_complete():
            raise RuntimeError("Cannot evaluate operation result before execution")

    @property
    def executed(self) -> bool:
        if self._commands is None:
            return False
        return len(self._commands) > 0

    @property
    def will_change(self) -> bool:
        if self._maybe_is_change is not None:
            return self._maybe_is_change

        op_data = context.state.get_op_data_for_host(context.host, self._hash)
        cmd_gen = op_data.command_generator
        for _ in cmd_gen():
            self._maybe_is_change = True
            return True
        self._maybe_is_change = False
        return False

    def did_change(self) -> bool:
        self._raise_if_not_complete()
        return bool(self._success and len(self._commands or []) > 0)

    def did_not_change(self) -> bool:
        return not self.did_change()

    def did_succeed(self, _raise_if_not_complete=True) -> bool:
        if _raise_if_not_complete:
            self._raise_if_not_complete()
        return self._success is True

    def did_error(self) -> bool:
        self._raise_if_not_complete()
        return self._success is False

    # TODO: deprecated, remove in v4
    @property
    def changed(self) -> bool:
        if self.is_complete():
            return self.did_change()
        return self.will_change

    @property
    def stdout_lines(self) -> list[str]:
        self._raise_if_not_complete()
        assert self._combined_output is not None
        return self._combined_output.stdout_lines

    @property
    def stderr_lines(self) -> list[str]:
        self._raise_if_not_complete()
        assert self._combined_output is not None
        return self._combined_output.stderr_lines

    @property
    def stdout(self) -> str:
        return "\n".join(self.stdout_lines)

    @property
    def stderr(self) -> str:
        return "\n".join(self.stderr_lines)


def add_op(state: State, op_func, *args, **kwargs):
    """
    Prepare & add an operation to ``pyinfra.state`` by executing it on all hosts.

    Args:
        state (``pyinfra.api.State`` obj): the deploy state to add the operation
        to op_func (function): the operation function from one of the modules,
        ie ``server.user``
        args/kwargs: passed to the operation function
    """

    if pyinfra.is_cli:
        raise PyinfraError(
            ("`add_op` should not be called when pyinfra is executing in CLI mode! ({0})").format(
                get_call_location(),
            ),
        )

    hosts = kwargs.pop("host", state.inventory.iter_active_hosts())
    if isinstance(hosts, Host):
        hosts = [hosts]

    with ctx_state.use(state):
        results = {}
        for op_host in hosts:
            with ctx_host.use(op_host):
                results[op_host] = op_func(*args, **kwargs)

    return results


P = ParamSpec("P")


def operation(
    is_idempotent: bool = True,
    idempotent_notice: Optional[str] = None,
    is_deprecated: bool = False,
    deprecated_for: Optional[str] = None,
    _set_in_op: bool = True,
) -> Callable[[Callable[P, Generator]], PyinfraOperation[P]]:
    """
    Decorator that takes a simple module function and turn it into the internal
    operation representation that consists of a list of commands + options
    (sudo, (sudo|su)_user, env).
    """

    def decorator(f: Callable[P, Generator]) -> PyinfraOperation[P]:
        f.is_idempotent = is_idempotent  # type: ignore[attr-defined]
        f.idempotent_notice = idempotent_notice  # type: ignore[attr-defined]
        f.is_deprecated = is_deprecated  # type: ignore[attr-defined]
        f.deprecated_for = deprecated_for  # type: ignore[attr-defined]
        return _wrap_operation(f, _set_in_op=_set_in_op)

    return decorator


def _wrap_operation(func: Callable[P, Generator], _set_in_op: bool = True) -> PyinfraOperation[P]:
    @wraps(func)
    def decorated_func(*args: P.args, **kwargs: P.kwargs) -> OperationMeta:
        state = context.state
        host = context.host

        if host.in_op:
            raise Exception(
                "Operation called within another operation, this is not allowed! Use the `_inner` "
                + "function to call the underlying operation."
            )

        if func.is_deprecated:  # type: ignore[attr-defined]
            if func.deprecated_for:  # type: ignore[attr-defined]
                logger.warning(
                    f"The {get_operation_name_from_func(func)} operation is "
                    + f"deprecated, please use: {func.deprecated_for}",  # type: ignore[attr-defined] # noqa
                )
            else:
                logger.warning(f"The {get_operation_name_from_func(func)} operation is deprecated")

        # Configure operation
        #
        # Get the meta kwargs (globals that apply to all hosts)
        global_arguments, global_argument_keys = pop_global_arguments(kwargs)

        names, add_args = generate_operation_name(func, host, kwargs, global_arguments)
        op_order, op_hash = solve_operation_consistency(names, state, host)

        # Ensure shared (between servers) operation meta, mutates state
        op_meta = ensure_shared_op_meta(state, op_hash, op_order, global_arguments, names)

        # Attach normal args, if we're auto-naming this operation
        if add_args:
            op_meta = attach_args(op_meta, args, kwargs)

        # Check if we're actually running the operation on this host
        # Run once and we've already added meta for this op? Stop here.
        if op_meta.global_arguments["_run_once"]:
            has_run = False
            for ops in state.ops.values():
                if op_hash in ops:
                    has_run = True
                    break

            if has_run:
                return OperationMeta(op_hash, is_change=False)

        # Grab a reference to any *current* deploy data as this may change when
        # we later evaluate the operation at runtime.This means we put back the
        # expected deploy data.
        current_deploy_data = host.current_deploy_data

        # "Run" operation - here we make a generator that will yield out actual commands to execute
        # and, if we're diff-ing, we then iterate the generator now to determine if any changes
        # *would* be made based on the *current* remote state.

        def command_generator() -> Iterator[PyinfraCommand]:
            # Check global _if argument function and do nothing if returns False
            if state.is_executing:
                _ifs = global_arguments.get("_if")
                if isinstance(_ifs, list) and not all(_if() for _if in _ifs):
                    return
                elif callable(_ifs) and not _ifs():
                    return

            host.in_op = _set_in_op
            host.current_op_hash = op_hash
            host.current_op_global_arguments = global_arguments
            host.current_op_deploy_data = current_deploy_data

            try:
                for command in func(*args, **kwargs):
                    if isinstance(command, str):
                        command = StringCommand(command.strip())
                    yield command
            finally:
                host.in_op = False
                host.current_op_hash = None
                host.current_op_global_arguments = None
                host.current_op_deploy_data = None

        op_is_change = None
        if state.should_check_for_changes():
            op_is_change = False
            for _ in command_generator():
                op_is_change = True
                break
        else:
            # If not calling the op function to check for change we still want to ensure the args
            # are valid, so use Signature.bind to trigger any TypeError.
            signature(func).bind(*args, **kwargs)

        # Add host-specific operation data to state, this mutates state
        host_meta = state.get_meta_for_host(host)
        host_meta.ops += 1
        if op_is_change:
            host_meta.ops_change += 1
        else:
            host_meta.ops_no_change += 1

        operation_meta = OperationMeta(op_hash, op_is_change)

        # Add the server-relevant commands
        op_data = StateOperationHostData(command_generator, global_arguments, operation_meta)
        state.set_op_data_for_host(host, op_hash, op_data)

        # If we're already in the execution phase, execute this operation immediately
        if state.is_executing:
            execute_immediately(state, host, op_hash)

        # Return result meta for use in deploy scripts
        return operation_meta

    decorated_func._inner = func  # type: ignore[attr-defined]
    return cast(PyinfraOperation[P], decorated_func)


def get_operation_name_from_func(func):
    if func.__module__:
        module_bits = func.__module__.split(".")
        module_name = module_bits[-1]
        return "{0}.{1}".format(module_name, func.__name__)
    else:
        return func.__name__


def generate_operation_name(func, host, kwargs, global_arguments):
    # Generate an operation name if needed (Module/Operation format)
    name = global_arguments.get("name")
    add_args = False
    if name:
        names = {name}
    else:
        add_args = True
        name = get_operation_name_from_func(func)
        names = {name}

    if host.current_deploy_name:
        names = {"{0} | {1}".format(host.current_deploy_name, name) for name in names}

    return names, add_args


def solve_operation_consistency(names, state, host):
    # Operation order is used to tie-break available nodes in the operation DAG, in CLI mode
    # we use stack call order so this matches as defined by the user deploy code.
    if pyinfra.is_cli:
        op_order = get_operation_order_from_stack(state)
    # In API mode we just increase the order for each host
    else:
        op_order = [len(host.op_hash_order)]

    if host.loop_position:
        op_order.extend(host.loop_position)

    # Make a hash from the call stack lines
    op_hash = make_hash(op_order)

    # Avoid adding duplicates! This happens if an operation is called within
    # a loop - such that the filename/lineno/code _are_ the same, but the
    # arguments might be different. We just append an increasing number to
    # the op hash and also handle below with the op order.
    duplicate_op_count = 0
    while op_hash in host.op_hash_order:
        logger.debug("Duplicate hash ({0}) detected!".format(op_hash))
        op_hash = "{0}-{1}".format(op_hash, duplicate_op_count)
        duplicate_op_count += 1

    host.op_hash_order.append(op_hash)
    if duplicate_op_count:
        op_order.append(duplicate_op_count)

    op_order = tuple(op_order)
    logger.debug(f"Adding operation, {names}, opOrder={op_order}, opHash={op_hash}")
    return op_order, op_hash


# NOTE: this function mutates state.op_meta for this hash
def ensure_shared_op_meta(
    state: State,
    op_hash: str,
    op_order: tuple[int, ...],
    global_arguments: AllArguments,
    names: set[str],
):
    op_meta = state.op_meta.setdefault(op_hash, StateOperationMeta(op_order))

    for key in EXECUTION_KWARG_KEYS:
        global_value = global_arguments.pop(key)  # type: ignore[misc]
        op_meta_value = op_meta.global_arguments.get(key, op_meta_default)

        if op_meta_value is not op_meta_default and global_value != op_meta_value:
            raise OperationValueError("Cannot have different values for `{0}`.".format(key))

        op_meta.global_arguments[key] = global_value  # type: ignore[literal-required]

    # Add any new names to the set
    op_meta.names.update(names)

    return op_meta


def execute_immediately(state, host, op_hash):
    op_meta = state.get_op_meta(op_hash)
    op_data = state.get_op_data_for_host(host, op_hash)
    op_data.parent_op_hash = host.executing_op_hash
    log_operation_start(op_meta, op_types=["nested"], prefix="")
    run_host_op(state, host, op_hash)


def _get_arg_value(arg):
    if isinstance(arg, FunctionType):
        return arg.__name__
    if isinstance(arg, StringIO):
        return f"StringIO(hash={get_file_sha1(arg)})"
    return arg


def attach_args(op_meta, args, kwargs):
    for arg in args:
        if arg not in op_meta.args:
            op_meta.args.append(str(_get_arg_value(arg)))

    # Attach keyword args
    for key, value in kwargs.items():
        arg = "=".join((str(key), str(_get_arg_value(value))))
        if arg not in op_meta.args:
            op_meta.args.append(arg)

    return op_meta



"""
Manager Docker Containers, Volumes and Networks
"""

from pyinfra import host
from pyinfra.api import operation
from pyinfra.facts.docker import DockerContainer, DockerNetwork, DockerVolume

from .util.docker import handle_docker


@operation()
def container(
    container,
    image="",
    ports=None,
    networks=None,
    volumes=None,
    env_vars=None,
    pull_always=False,
    present=True,
    force=False,
    start=True,
):
    """
    Manage Docker containers

    + container: name to identify the container
    + image: container image and tag ex: nginx:alpine
    + networks: network list to attach on container
    + ports: port list to expose
    + volumes: volume list to map on container
    + env_vars: environment varible list to inject on container
    + pull_always: force image pull
    + force: remove a contaner with same name and create a new one
    + present: whether the container should be up and running
    + start: start or stop the container

    **Examples:**

    .. code:: python

        # Run a container
        docker.container(
            name="Deploy Nginx container",
            container="nginx",
            image="nginx:alpine",
            ports=["80:80"],
            present=True,
            force=True,
            networks=["proxy", "services"],
            volumes=["nginx_data:/usr/share/nginx/html"],
            pull_always=True,
        )

        # Stop a container
        docker.container(
            name="Stop Nginx container",
            container="nginx",
            start=False,
        )

        # Start a container
        docker.container(
            name="Start Nginx container",
            container="nginx",
            start=True,
        )
    """

    existent_container = host.get_fact(DockerContainer, object_id=container)

    if force:
        if existent_container:
            yield handle_docker(
                resource="container",
                command="remove",
                container=container,
            )

    if present:
        if not existent_container or force:
            yield handle_docker(
                resource="container",
                command="create",
                container=container,
                image=image,
                ports=ports,
                networks=networks,
                volumes=volumes,
                env_vars=env_vars,
                pull_always=pull_always,
                present=present,
                force=force,
                start=start,
            )

    if existent_container and start:
        if existent_container[0]["State"]["Status"] != "running":
            yield handle_docker(
                resource="container",
                command="start",
                container=container,
            )

    if existent_container and not start:
        if existent_container[0]["State"]["Status"] == "running":
            yield handle_docker(
                resource="container",
                command="stop",
                container=container,
            )

    if existent_container and not present:
        yield handle_docker(
            resource="container",
            command="remove",
            container=container,
        )


@operation(is_idempotent=False)
def image(image, present=True):
    """
    Manage Docker images

    + image: Image and tag ex: nginx:alpine
    + present: whether the Docker image should be exist

    **Examples:**

    .. code:: python

        # Pull a Docker image
        docker.image(
            name="Pull nginx image",
            image="nginx:alpine",
            present=True,
        )

        # Remove a Docker image
        docker.image(
            name="Remove nginx image",
            image:"nginx:image",
            present=False,
        )
    """

    if present:
        yield handle_docker(
            resource="image",
            command="pull",
            image=image,
        )

    else:
        yield handle_docker(
            resource="image",
            command="remove",
            image=image,
        )


@operation()
def volume(volume, driver="", labels=None, present=True):
    """
    Manage Docker volumes

    + volume: Volume name
    + driver: Docker volume storage driver
    + labels: Label list to attach in the volume
    + present: whether the Docker volume should exist

    **Examples:**

    .. code:: python

        # Create a Docker volume
        docker.volume(
            name="Create nginx volume",
            volume="nginx_data",
            present=True
        )
    """

    existent_volume = host.get_fact(DockerVolume, object_id=volume)

    if present:

        if existent_volume:
            host.noop("Volume alredy exist!")
            return

        yield handle_docker(
            resource="volume",
            command="create",
            volume=volume,
            driver=driver,
            labels=labels,
            present=present,
        )

    else:
        if existent_volume is None:
            host.noop("There is no {0} volume!".format(volume))
            return

        yield handle_docker(
            resource="volume",
            command="remove",
            volume=volume,
        )


@operation()
def network(
    network,
    driver="",
    gateway="",
    ip_range="",
    ipam_driver="",
    subnet="",
    scope="",
    opts=None,
    ipam_opts=None,
    labels=None,
    ingress=False,
    attachable=False,
    present=True,
):
    """
    Manage docker networks

    + network_name: Image name
    + driver: Container image and tag ex: nginx:alpine
    + gateway: IPv4 or IPv6 Gateway for the master subnet
    + ip_range: Allocate container ip from a sub-range
    + ipam_driver: IP Address Management Driver
    + subnet: Subnet in CIDR format that represents a network segment
    + scope: Control the network's scope
    + opts: Set driver specific options
    + ipam_opts: Set IPAM driver specific options
    + labels: Label list to attach in the network
    + ingress: Create swarm routing-mesh network
    + attachable: Enable manual container attachment
    + present: whether the Docker network should exist

    **Examples:**

    .. code:: python

        # Create Docker network
        docker.network(
            name="Create nginx network",
            network_name="nginx",
            attachable=True,
            present=True,
        )
    """
    existent_network = host.get_fact(DockerNetwork, object_id=network)

    if present:
        if existent_network:
            host.noop("Alredy exist a network with {0} name!".format(network))
            return

        yield handle_docker(
            resource="network",
            command="create",
            network=network,
            driver=driver,
            gateway=gateway,
            ip_range=ip_range,
            ipam_driver=ipam_driver,
            subnet=subnet,
            scope=scope,
            opts=opts,
            ipam_opts=ipam_opts,
            labels=labels,
            ingress=ingress,
            attachable=attachable,
            present=present,
        )

    else:
        if existent_network is None:
            host.noop("Ther is not network with {0} name!".format(network))
            return

        yield handle_docker(
            resource="network",
            command="create",
            network=network,
        )


@operation(is_idempotent=False)
def prune(
    all=False,
    volume=False,
    filter="",
):
    """
    Execute a docker system prune.

    + all: Remove all unused images not just dangling ones
    + volumes: Prune anonymous volumes
    + filter: Provide filter values (e.g. "label=<key>=<value>" or "until=24h")

    **Examples:**

    .. code:: python

        # Remove dangling images
        docker.prune(
            name="remove dangling images",
        )

        # Remove all images and volumes
        docker.prune(
            name="Remove all images and volumes",
            all=True,
            volumes=True,
        )

        # Remove images older than 90 days
        docker.prune(
            name="Remove unused older than 90 days",
            filter="until=2160h"
        )
    """

    yield handle_docker(
        resource="system",
        command="prune",
        all=all,
        volume=volume,
        filter=filter,
    )



from __future__ import annotations

import shlex
from inspect import getfullargspec
from string import Formatter
from typing import IO, TYPE_CHECKING, Callable, Union

import gevent
from typing_extensions import Unpack

from pyinfra.context import ctx_config, ctx_host

from .arguments import ConnectorArguments

if TYPE_CHECKING:
    from pyinfra.api.host import Host
    from pyinfra.api.state import State


def make_formatted_string_command(string: str, *args, **kwargs) -> "StringCommand":
    """
    Helper function that takes a shell command or script as a string, splits it
    using ``shlex.split`` and then formats each bit, returning a ``StringCommand``
    instance with each bit.

    Useful to enable string formatted commands/scripts, for example:

    .. code:: python

        curl_command = make_formatted_string_command(
            'curl -sSLf {0} -o {1}',
            QuoteString(src),
            QuoteString(dest),
        )
    """

    formatter = Formatter()
    string_bits = []

    for bit in shlex.split(string):
        for item in formatter.parse(bit):
            if item[0]:
                string_bits.append(item[0])
            if item[1]:
                value, _ = formatter.get_field(item[1], args, kwargs)
                string_bits.append(value)

    return StringCommand(*string_bits)


class MaskString(str):
    pass


class QuoteString:
    obj: Union[str, "StringCommand"]

    def __init__(self, obj: Union[str, "StringCommand"]):
        self.obj = obj

    def __repr__(self) -> str:
        return f"QuoteString({self.obj})"


class PyinfraCommand:
    connector_arguments: ConnectorArguments

    def __init__(self, **arguments: Unpack[ConnectorArguments]):
        self.connector_arguments = arguments

    def __eq__(self, other) -> bool:
        if isinstance(other, self.__class__) and repr(self) == repr(other):
            return True
        return False

    def execute(self, state: "State", host: "Host", connector_arguments: ConnectorArguments):
        raise NotImplementedError


class StringCommand(PyinfraCommand):
    def __init__(
        self,
        *bits,
        _separator=" ",
        **arguments: Unpack[ConnectorArguments],
    ):
        super().__init__(**arguments)
        self.bits = bits
        self.separator = _separator

    def __str__(self) -> str:
        return self.get_masked_value()

    def __repr__(self) -> str:
        return f"StringCommand({self.get_masked_value()})"

    def _get_all_bits(self, bit_accessor):
        all_bits = []

        for bit in self.bits:
            quote = False
            if isinstance(bit, QuoteString):
                quote = True
                bit = bit.obj

            if isinstance(bit, StringCommand):
                bit = bit_accessor(bit)

            if not isinstance(bit, str):
                bit = "{0}".format(bit)

            if quote:
                bit = shlex.quote(bit)

            all_bits.append(bit)

        return all_bits

    def get_raw_value(self) -> str:
        return self.separator.join(
            self._get_all_bits(
                lambda bit: bit.get_raw_value(),
            ),
        )

    def get_masked_value(self) -> str:
        return self.separator.join(
            [
                "***" if isinstance(bit, MaskString) else bit
                for bit in self._get_all_bits(lambda bit: bit.get_masked_value())
            ],
        )

    def execute(self, state: "State", host: "Host", connector_arguments: ConnectorArguments):
        connector_arguments.update(self.connector_arguments)

        return host.run_shell_command(
            self,
            print_output=state.print_output,
            print_input=state.print_input,
            **connector_arguments,
        )


class FileUploadCommand(PyinfraCommand):
    def __init__(
        self,
        src: str | IO,
        dest: str,
        remote_temp_filename=None,
        **kwargs: Unpack[ConnectorArguments],
    ):
        super().__init__(**kwargs)
        self.src = src
        self.dest = dest
        self.remote_temp_filename = remote_temp_filename

    def __repr__(self):
        return "FileUploadCommand({0}, {1})".format(self.src, self.dest)

    def execute(self, state: "State", host: "Host", connector_arguments: ConnectorArguments):
        connector_arguments.update(self.connector_arguments)

        return host.put_file(
            self.src,
            self.dest,
            remote_temp_filename=self.remote_temp_filename,
            print_output=state.print_output,
            print_input=state.print_input,
            **connector_arguments,
        )


class FileDownloadCommand(PyinfraCommand):
    def __init__(
        self,
        src: str,
        dest: str | IO,
        remote_temp_filename=None,
        **kwargs: Unpack[ConnectorArguments],
    ):
        super().__init__(**kwargs)
        self.src = src
        self.dest = dest
        self.remote_temp_filename = remote_temp_filename

    def __repr__(self):
        return "FileDownloadCommand({0}, {1})".format(self.src, self.dest)

    def execute(self, state: "State", host: "Host", connector_arguments: ConnectorArguments):
        connector_arguments.update(self.connector_arguments)

        return host.get_file(
            self.src,
            self.dest,
            remote_temp_filename=self.remote_temp_filename,
            print_output=state.print_output,
            print_input=state.print_input,
            **connector_arguments,
        )


class FunctionCommand(PyinfraCommand):
    def __init__(
        self,
        function: Callable,
        args,
        func_kwargs,
        **kwargs: Unpack[ConnectorArguments],
    ):
        super().__init__(**kwargs)
        self.function = function
        self.args = args
        self.kwargs = func_kwargs

    def __repr__(self):
        return "FunctionCommand({0}, {1}, {2})".format(
            self.function.__name__,
            self.args,
            self.kwargs,
        )

    def execute(self, state: "State", host: "Host", connector_arguments: ConnectorArguments):
        argspec = getfullargspec(self.function)
        if "state" in argspec.args and "host" in argspec.args:
            return self.function(state, host, *self.args, **self.kwargs)

        def execute_function():
            with ctx_config.use(state.config.copy()):
                with ctx_host.use(host):
                    self.function(*self.args, **self.kwargs)

        greenlet = gevent.spawn(execute_function)
        return greenlet.get()


class RsyncCommand(PyinfraCommand):
    def __init__(self, src: str, dest: str, flags, **kwargs: Unpack[ConnectorArguments]):
        super().__init__(**kwargs)
        self.src = src
        self.dest = dest
        self.flags = flags

    def __repr__(self):
        return "RsyncCommand({0}, {1}, {2})".format(self.src, self.dest, self.flags)

    def execute(self, state: "State", host: "Host", connector_arguments: ConnectorArguments):
        return host.rsync(
            self.src,
            self.dest,
            self.flags,
            print_output=state.print_output,
            print_input=state.print_input,
            **connector_arguments,
        )



from __future__ import annotations

import re

from pyinfra.api import FactBase


class FlatpakBaseFact(FactBase):
    abstract = True

    def requires_command(self, *args, **kwargs) -> str:
        return "flatpak"


class FlatpakPackage(FlatpakBaseFact):
    """
    Returns information for an installed flatpak package

    .. code:: python

        {
            "id": "org.signal.Signal",
            "ref": "app/org.signal.Signal/x86_64/stable",
            "version": "7.12.0"
        }
    """

    default = dict
    _regexes = {
        "id": "^[ ]+ID:[ ]+(.*)$",
        "ref": r"^[ ]+Ref:[ ]+(.*)$",
        "version": r"^[ ]+Version:[ ]+([\w\d.-]+).*$",
    }

    def command(self, package):
        return f"flatpak info {package}"

    def process(self, output):
        data = {}
        for line in output:
            for regex_name, regex in self._regexes.items():
                matches = re.match(regex, line)
                if matches:
                    data[regex_name] = matches.group(1)

        return data


class FlatpakPackages(FlatpakBaseFact):
    """
    Returns a list of installed flatpak packages:

    .. code:: python

        [
            "org.gnome.Platform",
            "org.kde.Platform",
            "org.kde.Sdk",
            "org.libreoffice.LibreOffice",
            "org.videolan.VLC"
        ]
    """

    default = list

    def command(self):
        return "flatpak list --columns=application"

    def process(self, output):
        return [flatpak for flatpak in output[1:]]



"""
The files facts provide information about the filesystem and it's contents on the target host.
"""

from __future__ import annotations

import re
import shlex
import stat
from datetime import datetime
from typing import TYPE_CHECKING, List, Optional, Tuple, Union

from typing_extensions import Literal, NotRequired, TypedDict

from pyinfra.api.command import QuoteString, make_formatted_string_command
from pyinfra.api.facts import FactBase
from pyinfra.api.util import try_int

LINUX_STAT_COMMAND = "stat -c 'user=%U group=%G mode=%A atime=%X mtime=%Y ctime=%Z size=%s %N'"
BSD_STAT_COMMAND = "stat -f 'user=%Su group=%Sg mode=%Sp atime=%a mtime=%m ctime=%c size=%z %N%SY'"

STAT_REGEX = (
    r"user=(.*) group=(.*) mode=(.*) "
    r"atime=([0-9]*) mtime=([0-9]*) ctime=([0-9]*) "
    r"size=([0-9]*) (.*)"
)

FLAG_TO_TYPE = {
    "b": "block",
    "c": "character",
    "d": "directory",
    "l": "link",
    "s": "socket",
    "p": "fifo",
    "-": "file",
}

# Each item is a map of character to permission octal to be combined, taken from stdlib:
# https://github.com/python/cpython/blob/c1c3be0f9dc414bfae9a5718451ca217751ac687/Lib/stat.py#L128-L154
CHAR_TO_PERMISSION = (
    # User
    {"r": stat.S_IRUSR},
    {"w": stat.S_IWUSR},
    {"x": stat.S_IXUSR, "S": stat.S_ISUID, "s": stat.S_IXUSR | stat.S_ISUID},
    # Group
    {"r": stat.S_IRGRP},
    {"w": stat.S_IWGRP},
    {"x": stat.S_IXGRP, "S": stat.S_ISGID, "s": stat.S_IXGRP | stat.S_ISGID},
    # Other
    {"r": stat.S_IROTH},
    {"w": stat.S_IWOTH},
    {"x": stat.S_IXOTH, "T": stat.S_ISVTX, "t": stat.S_IXOTH | stat.S_ISVTX},
)


def _parse_mode(mode: str) -> int:
    """
    Converts ls mode output (rwxrwxrwx) -> octal permission integer (755).
    """

    out = 0

    for i, char in enumerate(mode):
        for c, m in CHAR_TO_PERMISSION[i].items():
            if char == c:
                out |= m
                break

    return int(oct(out)[2:])


def _parse_datetime(value: str) -> Optional[datetime]:
    value = try_int(value)
    if isinstance(value, int):
        return datetime.utcfromtimestamp(value)
    return None


class FileDict(TypedDict):
    mode: int
    size: Union[int, str]
    atime: Optional[datetime]
    mtime: Optional[datetime]
    ctime: Optional[datetime]
    user: str
    group: str
    link_target: NotRequired[str]


class File(FactBase[Union[FileDict, Literal[False], None]]):
    """
    Returns information about a file on the remote system:

    .. code:: python

        {
            "user": "pyinfra",
            "group": "pyinfra",
            "mode": 644,
            "size": 3928,
        }

    If the path does not exist:
        returns ``None``

    If the path exists but is not a file:
        returns ``False``
    """

    type = "file"

    def command(self, path):
        if path.startswith("~/"):
            # Do not quote leading tilde to ensure that it gets properly expanded by the shell
            path = f"~/{shlex.quote(path[2:])}"
        else:
            path = QuoteString(path)

        return make_formatted_string_command(
            (
                # only stat if the path exists (file or symlink)
                "! (test -e {0} || test -L {0} ) || "
                "( {linux_stat_command} {0} 2> /dev/null || {bsd_stat_command} {0} )"
            ),
            path,
            linux_stat_command=LINUX_STAT_COMMAND,
            bsd_stat_command=BSD_STAT_COMMAND,
        )

    def process(self, output) -> Union[FileDict, Literal[False], None]:
        match = re.match(STAT_REGEX, output[0])
        if not match:
            return None

        mode = match.group(3)
        path_type = FLAG_TO_TYPE[mode[0]]

        data: FileDict = {
            "user": match.group(1),
            "group": match.group(2),
            "mode": _parse_mode(mode[1:]),
            "atime": _parse_datetime(match.group(4)),
            "mtime": _parse_datetime(match.group(5)),
            "ctime": _parse_datetime(match.group(6)),
            "size": try_int(match.group(7)),
        }

        if path_type != self.type:
            return False

        if path_type == "link":
            filename = match.group(8)
            filename, target = filename.split(" -> ")
            data["link_target"] = target.strip("'").lstrip("`")

        return data


class Link(File):
    """
    Returns information about a link on the remote system:

    .. code:: python

        {
            "user": "pyinfra",
            "group": "pyinfra",
            "link_target": "/path/to/link/target"
        }

    If the path does not exist:
        returns ``None``

    If the path exists but is not a link:
        returns ``False``
    """

    type = "link"


class Directory(File):
    """
    Returns information about a directory on the remote system:

    .. code:: python

        {
            "user": "pyinfra",
            "group": "pyinfra",
            "mode": 644,
        }

    If the path does not exist:
        returns ``None``

    If the path exists but is not a directory:
        returns ``False``
    """

    type = "directory"


class Socket(File):
    """
    Returns information about a socket on the remote system:

    .. code:: python

        {
            "user": "pyinfra",
            "group": "pyinfra",
        }

    If the path does not exist:
        returns ``None``

    If the path exists but is not a socket:
        returns ``False``
    """

    type = "socket"


if TYPE_CHECKING:
    FactBaseOptionalStr = FactBase[Optional[str]]
else:
    FactBaseOptionalStr = FactBase


class HashFileFactBase(FactBaseOptionalStr):
    _raw_cmd: str
    _regexes: Tuple[str, str]

    def __init_subclass__(cls, digits: int, cmds: List[str], **kwargs) -> None:
        super().__init_subclass__(**kwargs)

        raw_hash_cmds = ["%s {0} 2> /dev/null" % cmd for cmd in cmds]
        raw_hash_cmd = " || ".join(raw_hash_cmds)
        cls._raw_cmd = "test -e {0} && ( %s ) || true" % raw_hash_cmd

        assert cls.__name__.endswith("File")
        hash_name = cls.__name__[:-4].upper()
        cls._regexes = (
            # GNU coreutils style:
            r"^([a-fA-F0-9]{%d})\s+%%s$" % digits,
            # BSD style:
            r"^%s\s+\(%%s\)\s+=\s+([a-fA-F0-9]{%d})$" % (hash_name, digits),
        )

    def command(self, path):
        self.path = path
        return make_formatted_string_command(self._raw_cmd, QuoteString(path))

    def process(self, output) -> Optional[str]:
        output = output[0]
        escaped_path = re.escape(self.path)
        for regex in self._regexes:
            matches = re.match(regex % escaped_path, output)
            if matches:
                return matches.group(1)
        return None


class Sha1File(HashFileFactBase, digits=40, cmds=["sha1sum", "shasum", "sha1"]):
    """
    Returns a SHA1 hash of a file. Works with both sha1sum and sha1. Returns
    ``None`` if the file doest not exist.
    """


class Sha256File(HashFileFactBase, digits=64, cmds=["sha256sum", "shasum -a 256", "sha256"]):
    """
    Returns a SHA256 hash of a file, or ``None`` if the file does not exist.
    """


class Md5File(HashFileFactBase, digits=32, cmds=["md5sum", "md5"]):
    """
    Returns an MD5 hash of a file, or ``None`` if the file does not exist.
    """


class FindInFile(FactBase):
    """
    Checks for the existence of text in a file using grep. Returns a list of matching
    lines if the file exists, and ``None`` if the file does not.
    """

    def command(self, path, pattern, interpolate_variables=False):
        self.exists_flag = "__pyinfra_exists_{0}".format(path)

        if interpolate_variables:
            pattern = '"{0}"'.format(pattern.replace('"', '\\"'))
        else:
            pattern = QuoteString(pattern)

        return make_formatted_string_command(
            (
                "grep -e {0} {1} 2> /dev/null || "
                "( find {1} -type f > /dev/null && echo {2} || true )"
            ),
            pattern,
            QuoteString(path),
            QuoteString(self.exists_flag),
        )

    def process(self, output):
        # If output is the special string: no matches, so return an empty list;
        # this allows us to differentiate between no matches in an existing file
        # or a file not existing.
        if output and output[0] == self.exists_flag:
            return []

        return output


class FindFilesBase(FactBase):
    abstract = True
    default = list
    type_flag: str

    def process(self, output):
        return output

    def command(self, path, quote_path=True):
        return make_formatted_string_command(
            "find {0} -type {type_flag} || true",
            QuoteString(path) if quote_path else path,
            type_flag=self.type_flag,
        )


class FindFiles(FindFilesBase):
    """
    Returns a list of files from a start path, recursively using ``find``.
    """

    type_flag = "f"


class FindLinks(FindFilesBase):
    """
    Returns a list of links from a start path, recursively using ``find``.
    """

    type_flag = "l"


class FindDirectories(FindFilesBase):
    """
    Returns a list of directories from a start path, recursively using ``find``.
    """

    type_flag = "d"


class Flags(FactBase):
    """
    Returns a list of the file flags set for the specified file or directory.
    """

    def requires_command(self, path) -> str:
        return "chflags"  # don't try to retrieve them if we can't set them

    def command(self, path):
        return make_formatted_string_command(
            "! test -e {0} || stat -f %Sf {0}",
            QuoteString(path),
        )

    def process(self, output):
        return [flag for flag in output[0].split(",") if len(flag) > 0] if len(output) == 1 else []


MARKER_DEFAULT = "# {mark} PYINFRA BLOCK"
MARKER_BEGIN_DEFAULT = "BEGIN"
MARKER_END_DEFAULT = "END"
EXISTS = "__pyinfra_exists_"
MISSING = "__pyinfra_missing_"


class Block(FactBase):
    """
    Returns a (possibly empty) list of the lines found between the markers.

    .. code:: python

        [
            "xray: one",
            "alpha: two"
        ]

    If the ``path`` doesn't exist
        returns ``None``

    If the ``path`` exists but the markers are not found
        returns ``[]``
    """

    # if markers aren't found, awk will return 0 and produce no output but we need to
    # distinguish between "markers not found" and "markers found but nothing between them"
    # for the former we use the empty list (created the call to default) and for the latter
    # the list with a single empty string.
    default = list

    def command(self, path, marker=None, begin=None, end=None):
        self.path = path
        start = (marker or MARKER_DEFAULT).format(mark=begin or MARKER_BEGIN_DEFAULT)
        end = (marker or MARKER_DEFAULT).format(mark=end or MARKER_END_DEFAULT)
        if start == end:
            raise ValueError(f"delimiters for block must be different but found only '{start}'")

        backstop = make_formatted_string_command(
            "(find {0} -type f > /dev/null && echo {1} || echo {2} )",
            QuoteString(path),
            QuoteString(f"{EXISTS}{path}"),
            QuoteString(f"{MISSING}{path}"),
        )
        # m_f_s_c inserts blanks in unfortunate places, e.g. after first slash
        cmd = make_formatted_string_command(
            f"awk \\'/{end}/{{{{f=0}}}} f; /{start}/{{{{f=1}}}}\\' {{0}} || {backstop}",
            QuoteString(path),
        )
        return cmd

    def process(self, output):
        if output and (output[0] == f"{EXISTS}{self.path}"):
            return []
        if output and (output[0] == f"{MISSING}{self.path}"):
            return None
        return output



from __future__ import annotations

import abc
from io import IOBase
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    Iterator,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
    get_type_hints,
)

from typing_extensions import TypedDict, Unpack

from pyinfra.api.exceptions import ConnectorDataTypeError
from pyinfra.api.util import raise_if_bad_type

if TYPE_CHECKING:
    from pyinfra.api.arguments import ConnectorArguments
    from pyinfra.api.command import StringCommand
    from pyinfra.api.host import Host, HostData
    from pyinfra.api.state import State

    from .util import CommandOutput


T = TypeVar("T")
default_sentinel = object()


def host_to_connector_data(
    connector_data: Type[T],
    connector_data_meta: dict[str, DataMeta],
    host_data: "HostData",
) -> T:
    data: T = cast(T, {})
    for key, type_ in get_type_hints(connector_data).items():
        value = host_data.get(key, default_sentinel)
        if value is default_sentinel:
            value = connector_data_meta[key].default
        else:
            raise_if_bad_type(
                value,
                type_,
                ConnectorDataTypeError,
                f"Invalid connector data `{key}`:",
            )

        data[key] = value  # type: ignore
    return data


class DataMeta:
    description: str
    default: Any

    def __init__(self, description, default=None) -> None:
        self.description = description
        self.default = default


class ConnectorData(TypedDict, total=False):
    pass


class BaseConnector(abc.ABC):
    state: "State"
    host: "Host"

    handles_execution = False

    data_cls: Type = ConnectorData
    data_meta: dict[str, DataMeta] = {}

    def __init__(self, state: "State", host: "Host"):
        self.state = state
        self.host = host
        self.data = host_to_connector_data(self.data_cls, self.data_meta, host.data)

    @staticmethod
    @abc.abstractmethod
    def make_names_data(name: str) -> Iterator[tuple[str, dict, list[str]]]:
        """
        Generates hosts/data/groups information for inventory. This allows a
        single connector reference to generate multiple target hosts.
        """
        ...

    def connect(self) -> None:
        """
        Connect this connector instance.
        """

    def disconnect(self) -> None:
        """
        Disconnect this connector instance.
        """

    @abc.abstractmethod
    def run_shell_command(
        self,
        command: "StringCommand",
        print_output: bool,
        print_input: bool,
        **arguments: Unpack["ConnectorArguments"],
    ) -> tuple[bool, "CommandOutput"]: ...

    @abc.abstractmethod
    def put_file(
        self,
        filename_or_io: Union[str, IOBase],
        remote_filename: str,
        remote_temp_filename: Optional[str] = None,
        print_output: bool = False,
        print_input: bool = False,
        **arguments: Unpack["ConnectorArguments"],
    ) -> bool: ...

    @abc.abstractmethod
    def get_file(
        self,
        remote_filename: str,
        filename_or_io: Union[str, IOBase],
        remote_temp_filename: Optional[str] = None,
        print_output: bool = False,
        print_input: bool = False,
        **arguments: Unpack["ConnectorArguments"],
    ) -> bool: ...

    def check_can_rsync(self):
        raise NotImplementedError("This connector does not support rsync")

    def rsync(
        self,
        src: str,
        dest: str,
        flags: Iterable[str],
        print_output: bool = False,
        print_input: bool = False,
        **arguments: Unpack["ConnectorArguments"],
    ) -> bool:
        raise NotImplementedError("This connector does not support rsync")



from __future__ import annotations

from pyinfra.api import FactBase

from .util.packaging import parse_packages

PKGIN_REGEX = r"^([a-zA-Z\-0-9]+)-([0-9\.]+\-?[a-z0-9]*)\s"


class PkginPackages(FactBase):
    """
    Returns a dict of installed pkgin packages:

    .. code:: python

        {
            "package_name": ["version"],
        }
    """

    def command(self) -> str:
        return "pkgin list"

    def requires_command(self) -> str:
        return "pkgin"

    default = dict

    def process(self, output):
        return parse_packages(PKGIN_REGEX, output)



from __future__ import annotations

import re

from pyinfra.api.facts import FactBase


class GitFactBase(FactBase):
    def requires_command(self, *args, **kwargs) -> str:
        return "git"


class GitBranch(GitFactBase):
    def command(self, repo) -> str:
        return "! test -d {0} || (cd {0} && git describe --all)".format(repo)

    def process(self, output):
        return re.sub(r"(heads|tags)/", r"", "\n".join(output))


class GitConfig(GitFactBase):
    default = dict

    def command(self, repo=None) -> str:
        if repo is None:
            return "git config --global -l || true"

        return "! test -d {0} || (cd {0} && git config --local -l)".format(repo)

    def process(self, output):
        items: dict[str, list[str]] = {}

        for line in output:
            key, value = line.split("=", 1)
            items.setdefault(key, []).append(value)

        return items


class GitTrackingBranch(GitFactBase):
    def command(self, repo) -> str:
        return r"! test -d {0} || (cd {0} && git status --branch --porcelain)".format(repo)

    def process(self, output):
        if not output:
            return None

        m = re.search(r"\.{3}(\S+)\b", list(output)[0])
        if m:
            return m.group(1)
        return None



from __future__ import annotations

import shlex
from collections import defaultdict
from io import StringIO
from typing import Callable
from urllib.parse import urlparse

from pyinfra.api import Host, State
from pyinfra.facts.files import File
from pyinfra.facts.rpm import RpmPackage
from pyinfra.operations import files


def _package_name(package: list[str] | str) -> str:
    if isinstance(package, list):
        return package[0]
    return package


def _has_package(
    package: str | list[str],
    packages: dict[str, set[str]],
    expand_package_fact: Callable[[str], list[str | list[str]]] | None = None,
    match_any=False,
) -> tuple[bool, dict]:
    def in_packages(pkg_name, pkg_versions):
        if not pkg_versions:
            return pkg_name in packages
        return pkg_name in packages and any(
            version in packages[pkg_name] for version in pkg_versions
        )

    packages_to_check: list[str | list[str]] = [package]
    if expand_package_fact:
        if isinstance(package, list):
            packages_to_check = expand_package_fact(package[0]) or packages_to_check
        else:
            packages_to_check = expand_package_fact(package) or packages_to_check

    package_name_to_versions = defaultdict(set)
    for pkg in packages_to_check:
        if isinstance(pkg, list):
            package_name_to_versions[pkg[0]].add(pkg[1])
        else:
            package_name_to_versions[pkg]  # just make sure it exists

    checks = (
        in_packages(pkg_name, pkg_versions)
        for pkg_name, pkg_versions in package_name_to_versions.items()
    )

    if match_any:
        return any(checks), package_name_to_versions
    return all(checks), package_name_to_versions


def ensure_packages(
    host: Host,
    packages_to_ensure: str | list[str] | None,
    current_packages: dict[str, set[str]],
    present: bool,
    install_command: str,
    uninstall_command: str,
    latest=False,
    upgrade_command: str | None = None,
    version_join: str | None = None,
    expand_package_fact: Callable[[str], list[str | list[str]]] | None = None,
):
    """
    Handles this common scenario:

    + We have a list of packages(/versions) to ensure
    + We have a map of existing package -> versions
    + We have the common command bits (install, uninstall, version "joiner")
    + Outputs commands to ensure our desired packages/versions
    + Optionally upgrades packages w/o specified version when present

    Args:
        packages_to_ensure (list): list of packages or package/versions
        current_packages (fact): fact returning dict of package names -> version
        present (bool): whether packages should exist or not
        install_command (str): command to prefix to list of packages to install
        uninstall_command (str): as above for uninstalling packages
        latest (bool): whether to upgrade installed packages when present
        upgrade_command (str): as above for upgrading
        version_join (str): the package manager specific "joiner", ie ``=`` for \
            ``<apt_pkg>=<version>``
        expand_package_fact: fact returning packages providing a capability \
            (ie ``yum whatprovides``)
    """

    if packages_to_ensure is None:
        return
    if isinstance(packages_to_ensure, str):
        packages_to_ensure = [packages_to_ensure]

    packages: list[str | list[str]] = packages_to_ensure  # type: ignore[assignment]

    if version_join:
        packages = [
            package[0] if len(package) == 1 else package
            for package in [package.rsplit(version_join, 1) for package in packages]  # type: ignore[union-attr] # noqa
        ]

    diff_packages = []
    diff_expanded_packages = {}

    upgrade_packages = []

    if present is True:
        for package in packages:
            has_package, expanded_packages = _has_package(
                package,
                current_packages,
                expand_package_fact,
            )

            if not has_package:
                diff_packages.append(package)
                diff_expanded_packages[_package_name(package)] = expanded_packages
            else:
                # Present packages w/o version specified - for upgrade if latest
                if isinstance(package, str):
                    upgrade_packages.append(package)

                if not latest:
                    pkg_name = _package_name(package)
                    if pkg_name in current_packages:
                        host.noop(
                            "package {0} is installed ({1})".format(
                                package,
                                ", ".join(current_packages[pkg_name]),
                            ),
                        )
                    else:
                        host.noop("package {0} is installed".format(package))

    if present is False:
        for package in packages:
            # String version, just check if existing
            has_package, expanded_packages = _has_package(
                package,
                current_packages,
                expand_package_fact,
                match_any=True,
            )

            if has_package:
                diff_packages.append(package)
                diff_expanded_packages[_package_name(package)] = expanded_packages
            else:
                host.noop("package {0} is not installed".format(package))

    if diff_packages:
        command = install_command if present else uninstall_command

        joined_packages = [
            version_join.join(package) if isinstance(package, list) else package  # type: ignore[union-attr] # noqa
            for package in diff_packages
        ]

        yield "{0} {1}".format(
            command,
            " ".join([shlex.quote(pkg) for pkg in joined_packages]),
        )

    if latest and upgrade_command and upgrade_packages:
        yield "{0} {1}".format(
            upgrade_command,
            " ".join([shlex.quote(pkg) for pkg in upgrade_packages]),
        )


def ensure_rpm(state: State, host: Host, source: str, present: bool, package_manager_command: str):
    original_source = source

    # If source is a url
    if urlparse(source).scheme:
        # Generate a temp filename (with .rpm extension to please yum)
        temp_filename = "{0}.rpm".format(host.get_temp_filename(source))

        # Ensure it's downloaded
        yield from files.download._inner(src=source, dest=temp_filename)

        # Override the source with the downloaded file
        source = temp_filename

    # Check for file .rpm information
    info = host.get_fact(RpmPackage, package=source)
    exists = False

    # We have info!
    if info:
        current_package = host.get_fact(RpmPackage, package=info["name"])
        if current_package and current_package["version"] == info["version"]:
            exists = True

    # Package does not exist and we want?
    if present and not exists:
        # If we had info, always install
        if info:
            yield "rpm -i {0}".format(source)
        # This happens if we download the package mid-deploy, so we have no info
        # but also don't know if it's installed. So check at runtime, otherwise
        # the install will fail.
        else:
            yield "rpm -q `rpm -qp {0}` 2> /dev/null || rpm -i {0}".format(source)

    # Package exists but we don't want?
    elif exists and not present:
        yield "{0} remove -y {1}".format(package_manager_command, info["name"])
    else:
        host.noop(
            "rpm {0} is {1}".format(
                original_source,
                "installed" if present else "not installed",
            ),
        )


def ensure_yum_repo(
    host: Host,
    name_or_url: str,
    baseurl: str | None,
    present: bool,
    description: str | None,
    enabled: bool,
    gpgcheck: bool,
    gpgkey: str | None,
    repo_directory="/etc/yum.repos.d/",
    type_: str | None = None,
):
    url = None
    url_parts = urlparse(name_or_url)
    if url_parts.scheme:
        url = name_or_url
        name_or_url = url_parts.path.split("/")[-1]
        if name_or_url.endswith(".repo"):
            name_or_url = name_or_url[:-5]

    filename = "{0}{1}.repo".format(repo_directory, name_or_url)

    # If we don't want the repo, just remove any existing file
    if not present:
        yield from files.file._inner(path=filename, present=False)
        return

    # If we're a URL, download the repo if it doesn't exist
    if url:
        if not host.get_fact(File, path=filename):
            yield from files.download._inner(src=url, dest=filename)
        return

    assert isinstance(baseurl, str)

    # Description defaults to name
    description = description or name_or_url

    # Build the repo file from string
    repo_lines = [
        "[{0}]".format(name_or_url),
        "name={0}".format(description),
        "baseurl={0}".format(baseurl),
        "enabled={0}".format(1 if enabled else 0),
        "gpgcheck={0}".format(1 if gpgcheck else 0),
    ]

    if type_:
        repo_lines.append("type={0}".format(type_))

    if gpgkey:
        repo_lines.append("gpgkey={0}".format(gpgkey))

    repo_lines.append("")
    repo = "\n".join(repo_lines)
    repo_file = StringIO(repo)

    # Ensure this is the file on the server
    yield from files.put._inner(src=repo_file, dest=filename)



"""
Deploys come in two forms: on-disk, eg deploy.py, and @deploy wrapped functions.
The latter enable re-usable (across CLI and API based execution) pyinfra extension
creation (eg pyinfra-openstack).
"""

from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, Optional, cast

from typing_extensions import ParamSpec

import pyinfra
from pyinfra import context
from pyinfra.context import ctx_host, ctx_state

from .arguments import pop_global_arguments
from .arguments_typed import PyinfraOperation
from .exceptions import PyinfraError
from .host import Host
from .util import get_call_location

if TYPE_CHECKING:
    from pyinfra.api.state import State


def add_deploy(state: "State", deploy_func: Callable[..., Any], *args, **kwargs):
    """
    Prepare & add an deploy to pyinfra.state by executing it on all hosts.

    Args:
        state (``pyinfra.api.State`` obj): the deploy state to add the operation
        deploy_func (function): the operation function from one of the modules,
        ie ``server.user``
        args/kwargs: passed to the operation function
    """

    if pyinfra.is_cli:
        raise PyinfraError(
            (
                "`add_deploy` should not be called when pyinfra is executing in CLI mode! ({0})"
            ).format(get_call_location()),
        )

    hosts = kwargs.pop("host", state.inventory.iter_active_hosts())
    if isinstance(hosts, Host):
        hosts = [hosts]

    with ctx_state.use(state):
        for deploy_host in hosts:
            with ctx_host.use(deploy_host):
                deploy_func(*args, **kwargs)


P = ParamSpec("P")


def deploy(name: Optional[str] = None, data_defaults=None):
    """
    Decorator that takes a deploy function (normally from a pyinfra_* package)
    and wraps any operations called inside with any deploy-wide kwargs/data.
    """

    if name and not isinstance(name, str):
        raise PyinfraError(
            (
                "The `deploy` decorator must be called, ie `@deploy()`, "
                "see: https://docs.pyinfra.com/en/3.x/compatibility.html#upgrading-pyinfra-from-2-x-3-x"  # noqa
            )
        )

    def decorator(func: Callable[P, Any]) -> PyinfraOperation[P]:
        func.deploy_name = name or func.__name__  # type: ignore[attr-defined]
        if data_defaults:
            func.deploy_data = data_defaults  # type: ignore[attr-defined]
        return _wrap_deploy(func)

    return decorator


def _wrap_deploy(func: Callable[P, Any]) -> PyinfraOperation[P]:
    @wraps(func)
    def decorated_func(*args: P.args, **kwargs: P.kwargs) -> Any:
        deploy_kwargs, _ = pop_global_arguments(kwargs)

        deploy_data = getattr(func, "deploy_data", None)

        with context.host.deploy(
            name=func.deploy_name,  # type: ignore[attr-defined]
            kwargs=deploy_kwargs,
            data=deploy_data,
        ):
            return func(*args, **kwargs)

    decorated_func._inner = func  # type: ignore[attr-defined]
    return cast(PyinfraOperation[P], decorated_func)



from __future__ import annotations

import json
import os
from datetime import datetime
from importlib import import_module
from importlib.util import find_spec
from io import IOBase
from os import path
from pathlib import Path
from types import CodeType, FunctionType, ModuleType
from typing import Callable

import click
import gevent

from pyinfra import logger, state
from pyinfra.api.command import PyinfraCommand
from pyinfra.api.exceptions import PyinfraError
from pyinfra.api.host import HostData
from pyinfra.api.operation import OperationMeta
from pyinfra.api.state import (
    State,
    StateHostMeta,
    StateHostResults,
    StateOperationHostData,
    StateOperationMeta,
)
from pyinfra.context import ctx_config, ctx_host
from pyinfra.progress import progress_spinner

from .exceptions import CliError, UnexpectedExternalError

# Cache for compiled Python deploy code
PYTHON_CODES: dict[str, CodeType] = {}


def is_subdir(child, parent):
    child = path.realpath(child)
    parent = path.realpath(parent)
    relative = path.relpath(child, start=parent)
    return not relative.startswith(os.pardir)


def exec_file(filename, return_locals: bool = False, is_deploy_code: bool = False):
    """
    Execute a Python file and optionally return it's attributes as a dict.
    """

    old_current_exec_filename = state.current_exec_filename
    state.current_exec_filename = filename

    if filename not in PYTHON_CODES:
        with open(filename, "r", encoding="utf-8") as f:
            code_str = f.read()

        code = compile(code_str, filename, "exec")
        PYTHON_CODES[filename] = code

    # Create some base attributes for our "module"
    data = {"__file__": filename}

    # Execute the code with locals/globals going into the dict above
    try:
        exec(PYTHON_CODES[filename], data)
    except PyinfraError:
        # Raise pyinfra errors as-is
        raise
    except Exception as e:
        # Wrap & re-raise errors in user code so we highlight filename/etc
        raise UnexpectedExternalError(e, filename)
    finally:
        state.current_exec_filename = old_current_exec_filename

    return data


def json_encode(obj):
    # pyinfra types
    if isinstance(obj, HostData):
        return obj.dict()

    if isinstance(obj, PyinfraCommand):
        return repr(obj)

    if isinstance(
        obj,
        (
            OperationMeta,
            StateOperationMeta,
            StateOperationHostData,
            StateHostMeta,
            StateHostResults,
        ),
    ):
        return repr(obj)

    # Python types
    if isinstance(obj, ModuleType):
        return "Module: {0}".format(obj.__name__)

    if isinstance(obj, FunctionType):
        return "Function: {0}".format(obj.__name__)

    if isinstance(obj, datetime):
        return obj.isoformat()

    if isinstance(obj, IOBase):
        if hasattr(obj, "name"):
            return "File: {0}".format(obj.name)

        if hasattr(obj, "template"):
            return "Template: {0}".format(obj.template)

        obj.seek(0)
        return "In memory file: {0}".format(obj.read())

    if isinstance(obj, Path):
        return str(obj)

    if isinstance(obj, set):
        return sorted(list(obj))

    if isinstance(obj, bytes):
        return obj.decode()

    raise TypeError("Cannot serialize: {0} ({1})".format(type(obj), obj))


def parse_cli_arg(arg):
    if isinstance(arg, list):
        return [parse_cli_arg(a) for a in arg]

    if arg.lower() == "false":
        return False

    if arg.lower() == "true":
        return True

    try:
        return int(arg)
    except (TypeError, ValueError):
        pass

    try:
        return json.loads(arg)
    except ValueError:
        pass

    return arg


def try_import_module_attribute(path, prefix=None, raise_for_none=True):
    if ":" in path:
        # Allow a.module.name:function syntax
        mod_path, attr_name = path.rsplit(":", 1)
    elif "." in path:
        # And also a.module.name.function
        mod_path, attr_name = path.rsplit(".", 1)
    else:
        return None

    possible_modules = [mod_path]
    if prefix:
        possible_modules.append(f"{prefix}.{mod_path}")

    module = None

    for possible in possible_modules:
        try:
            # First use find_spec which checks if the possible module exists *without* importing
            # it, thus any import errors it contains still get properly raised to the user.
            spec = find_spec(possible)
        except ModuleNotFoundError:
            continue
        else:
            if spec is not None:
                module = import_module(possible)
                break

    if module is None:
        if raise_for_none:
            raise CliError(f"No such module: {possible_modules[0]}")
        return

    attr = getattr(module, attr_name, None)
    if attr is None:
        if raise_for_none:
            raise CliError(f"No such attribute in module {possible_modules[0]}: {attr_name}")
        return

    return attr


def _parallel_load_hosts(state: "State", callback: Callable, name: str):
    def load_file(local_host):
        try:
            with ctx_config.use(state.config.copy()):
                with ctx_host.use(local_host):
                    callback()
                    logger.info(
                        "{0}{1} {2}".format(
                            local_host.print_prefix,
                            click.style("Ready:", "green"),
                            click.style(name, bold=True),
                        ),
                    )
        except Exception as e:
            return e

    greenlet_to_host = {
        state.pool.spawn(load_file, host): host for host in state.inventory.iter_active_hosts()
    }

    with progress_spinner(greenlet_to_host.values()) as progress:
        for greenlet in gevent.iwait(greenlet_to_host.keys()):
            host = greenlet_to_host[greenlet]
            result = greenlet.get()
            if isinstance(result, Exception):
                raise result
            progress(host)


def load_deploy_file(state: "State", filename):
    state.current_deploy_filename = filename
    _parallel_load_hosts(state, lambda: exec_file(filename), filename)


def load_func(state: "State", op_func, *args, **kwargs):
    _parallel_load_hosts(state, lambda: op_func(*args, **kwargs), op_func.__name__)



"""
The server module takes care of os-level state. Targets POSIX compatibility, tested on
Linux/BSD.
"""

from __future__ import annotations

import shlex
from io import StringIO
from itertools import filterfalse, tee
from os import path
from time import sleep
from typing import TYPE_CHECKING

from pyinfra import host, logger, state
from pyinfra.api import FunctionCommand, OperationError, StringCommand, operation
from pyinfra.api.util import try_int
from pyinfra.connectors.util import remove_any_sudo_askpass_file
from pyinfra.facts.files import Directory, FindInFile, Link
from pyinfra.facts.server import (
    Crontab,
    Groups,
    Home,
    Hostname,
    KernelModules,
    Locales,
    Mounts,
    Os,
    Sysctl,
    Users,
    Which,
)

from . import (
    apk,
    apt,
    brew,
    bsdinit,
    dnf,
    files,
    openrc,
    pacman,
    pkg,
    runit,
    systemd,
    sysvinit,
    upstart,
    xbps,
    yum,
    zypper,
)
from .util.files import chmod, sed_replace

if TYPE_CHECKING:
    from pyinfra.api.arguments_typed import PyinfraOperation


@operation(is_idempotent=False)
def reboot(delay=10, interval=1, reboot_timeout=300):
    """
    Reboot the server and wait for reconnection.

    + delay: number of seconds to wait before attempting reconnect
    + interval: interval (s) between reconnect attempts
    + reboot_timeout: total time before giving up reconnecting

    **Example:**

    .. code:: python

        server.reboot(
            name="Reboot the server and wait to reconnect",
            delay=60,
            reboot_timeout=600,
        )
    """

    # Remove this now, before we reboot the server - if the reboot fails (expected or
    # not) we'll error if we don't clean this up now. Will simply be re-uploaded if
    # needed later.
    def remove_any_askpass_file(state, host):
        remove_any_sudo_askpass_file(host)

    yield FunctionCommand(remove_any_askpass_file, (), {})

    yield StringCommand("reboot", _success_exit_codes=[0, -1])  # -1 being error/disconnected

    def wait_and_reconnect(state, host):  # pragma: no cover
        sleep(delay)
        max_retries = round(reboot_timeout / interval)

        host.disconnect()  # make sure we are properly disconnected
        retries = 0

        while True:
            host.connect(show_errors=False)
            if host.connected:
                break

            if retries > max_retries:
                raise Exception(
                    ("Server did not reboot in time (reboot_timeout={0}s)").format(reboot_timeout),
                )

            sleep(interval)
            retries += 1

    yield FunctionCommand(wait_and_reconnect, (), {})

    # On certain systems sudo files are lost on reboot
    def clean_sudo_info(state, host):
        host.connector_data["sudo_askpass_path"] = None

    yield FunctionCommand(clean_sudo_info, (), {})


@operation(is_idempotent=False)
def wait(port: int):
    """
    Waits for a port to come active on the target machine. Requires netstat, checks every
    second.

    + port: port number to wait for

    **Example:**

    .. code:: python

        server.wait(
            name="Wait for webserver to start",
            port=80,
        )
    """

    yield r"""
        while ! (netstat -an | grep LISTEN | grep -e "\.{0}" -e ":{0}"); do
            echo "waiting for port {0}..."
            sleep 1
        done
    """.format(
        port,
    )


@operation(is_idempotent=False)
def shell(commands: str | list[str]):
    """
    Run raw shell code on server during a deploy. If the command would
    modify data that would be in a fact, the fact would not be updated
    since facts are only run at the start of a deploy.

    + commands: command or list of commands to execute on the remote server

    **Example:**

    .. code:: python

        server.shell(
            name="Run lxd auto init",
            commands=["lxd init --auto"],
        )
    """

    # Ensure we have a list
    if isinstance(commands, str):
        commands = [commands]

    for command in commands:
        yield command


@operation(is_idempotent=False)
def script(src: str, args=()):
    """
    Upload and execute a local script on the remote host.

    + src: local script filename to upload & execute
    + args: iterable to pass as arguments to the script

    **Example:**

    .. code:: python

        # Note: This assumes there is a file in files/hello.bash locally.
        server.script(
            name="Hello",
            src="files/hello.bash",
        )

        # Example passing arguments to the script
        server.script(
            name="Hello",
            src="files/hello.bash",
            args=("do-something", "with-this"),
        )
    """

    temp_file = host.get_temp_filename()
    yield from files.put._inner(src=src, dest=temp_file)

    yield chmod(temp_file, "+x")
    yield StringCommand(temp_file, *args)


@operation(is_idempotent=False)
def script_template(src: str, args=(), **data):
    """
    Generate, upload and execute a local script template on the remote host.

    + src: local script template filename

    **Example:**

    .. code:: python

        # Example showing how to pass python variable to a script template file.
        # The .j2 file can use `{{ some_var }}` to be interpolated.
        # To see output need to run pyinfra with '-v'
        # Note: This assumes there is a file in templates/hello2.bash.j2 locally.
        some_var = 'blah blah blah '
        server.script_template(
            name="Hello from script",
            src="templates/hello2.bash.j2",
            some_var=some_var,
        )
    """

    temp_file = host.get_temp_filename("{0}{1}".format(src, data))
    yield from files.template._inner(src, temp_file, **data)

    yield chmod(temp_file, "+x")
    yield StringCommand(temp_file, *args)


@operation()
def modprobe(module: str, present=True, force=False):
    """
    Load/unload kernel modules.

    + module: name of the module to manage
    + present: whether the module should be loaded or not
    + force: whether to force any add/remove modules

    **Example:**

    .. code:: python

        server.modprobe(
            name="Silly example for modprobe",
            module="floppy",
        )
    """
    list_value = [module] if isinstance(module, str) else module

    # NOTE: https://docs.python.org/3/library/itertools.html#itertools-recipes
    def partition(predicate, iterable):
        t1, t2 = tee(iterable)
        return list(filter(predicate, t2)), list(filterfalse(predicate, t1))

    modules = host.get_fact(KernelModules)
    present_mods, missing_mods = partition(lambda mod: mod in modules, list_value)

    args = ""
    if force:
        args = " -f"

    # Module is loaded and we don't want it?
    if not present and present_mods:
        yield "modprobe{0} -r -a {1}".format(args, " ".join(present_mods))

    # Module isn't loaded and we want it?
    elif present and missing_mods:
        yield "modprobe{0} -a {1}".format(args, " ".join(missing_mods))

    else:
        host.noop(
            "{0} {1} {2} {3}".format(
                "modules" if len(list_value) > 1 else "module",
                "/".join(list_value),
                "are" if len(list_value) > 1 else "is",
                "loaded" if present else "not loaded",
            ),
        )


@operation()
def mount(
    path: str,
    mounted=True,
    options: list[str] | None = None,
    device: str | None = None,
    fs_type: str | None = None,
    # TODO: do we want to manage fstab here?
    # update_fstab=False,
):
    """
    Manage mounted filesystems.

    + path: the path of the mounted filesystem
    + mounted: whether the filesystem should be mounted
    + options: the mount options

    Options:
        If the currently mounted filesystem does not have all of the provided
        options it will be remounted with the options provided.

    ``/etc/fstab``:
        This operation does not attempt to modify the on disk fstab file - for
        that you should use the `files.line operation <./files.html#files-line>`_.
    """
    options = options or []
    options_string = ",".join(options)

    mounts = host.get_fact(Mounts)
    is_mounted = path in mounts

    # Want mount but don't have?
    if mounted and not is_mounted:
        args = []
        if fs_type:
            args.extend(["-t", fs_type])
        if options_string:
            args.extend(["-o", options_string])
        if device:
            args.append(device)
        args.append(path)

        yield StringCommand("mount", *args)

    # Want no mount but mounted?
    elif mounted is False and is_mounted:
        yield "umount {0}".format(path)

    # Want mount and is mounted! Check the options
    elif is_mounted and mounted and options:
        mounted_options = mounts[path]["options"]
        needed_options = set(options) - set(mounted_options)
        if needed_options:
            yield "mount -o remount,{0} {1}".format(options_string, path)

    else:
        host.noop(
            "filesystem {0} is {1}".format(
                path,
                "mounted" if mounted else "not mounted",
            ),
        )


@operation()
def hostname(hostname: str, hostname_file: str | None = None):
    """
    Set the system hostname using ``hostnamectl`` or ``hostname`` on older systems.

    + hostname: the hostname that should be set
    + hostname_file: the file that permanently sets the hostname

    Hostname file:
        The hostname file only matters no systems that do not have ``hostnamectl``,
        which is part of ``systemd``.

        By default pyinfra will auto detect this by targeting ``/etc/hostname``
        on Linux and ``/etc/myname`` on OpenBSD.

        To completely disable writing the hostname file, set ``hostname_file=False``.

    **Example:**

    .. code:: python

        server.hostname(
            name="Set the hostname",
            hostname="server1.example.com",
        )
    """

    current_hostname = host.get_fact(Hostname)

    if host.get_fact(Which, command="hostnamectl"):
        if current_hostname != hostname:
            yield "hostnamectl set-hostname {0}".format(hostname)
        else:
            host.noop("hostname is set")
        return

    if hostname_file is None:
        os = host.get_fact(Os)

        if os == "Linux":
            hostname_file = "/etc/hostname"
        elif os == "OpenBSD":
            hostname_file = "/etc/myname"

    if current_hostname != hostname:
        yield "hostname {0}".format(hostname)
    else:
        host.noop("hostname is set")

    if hostname_file:
        # Create a whole new hostname file
        file = StringIO("{0}\n".format(hostname))

        # And ensure it exists
        yield from files.put._inner(src=file, dest=hostname_file)


@operation()
def sysctl(
    key: str,
    value: str | int | list[str | int],
    persist=False,
    persist_file="/etc/sysctl.conf",
):
    """
    Edit sysctl configuration.

    + key: name of the sysctl setting to ensure
    + value: the value or list of values the sysctl should be
    + persist: whether to write this sysctl to the config
    + persist_file: file to write the sysctl to persist on reboot

    **Example:**

    .. code:: python

        server.sysctl(
            name="Change the fs.file-max value",
            key="fs.file-max",
            value=100000,
            persist=True,
        )
    """

    string_value = " ".join(["{0}".format(v) for v in value]) if isinstance(value, list) else value

    value = [try_int(v) for v in value] if isinstance(value, list) else try_int(value)

    existing_sysctls = host.get_fact(Sysctl, keys=[key])
    existing_value = existing_sysctls.get(key)

    if not existing_value or existing_value != value:
        yield "sysctl {0}='{1}'".format(key, string_value)
    else:
        host.noop("sysctl {0} is set to {1}".format(key, string_value))

    if persist:
        yield from files.line._inner(
            path=persist_file,
            line="{0}[[:space:]]*=[[:space:]]*{1}".format(key, string_value),
            replace="{0} = {1}".format(key, string_value),
        )


@operation()
def service(
    service: str,
    running=True,
    restarted=False,
    reloaded=False,
    command: str | None = None,
    enabled: bool | None = None,
):
    """
    Manage the state of services. This command checks for the presence of all the
    Linux init systems pyinfra can handle and executes the relevant operation.

    + service: name of the service to manage
    + running: whether the service should be running
    + restarted: whether the service should be restarted
    + reloaded: whether the service should be reloaded
    + command: custom command execute
    + enabled: whether this service should be enabled/disabled on boot

    **Example:**

    .. code:: python

        server.service(
            name="Enable open-vm-tools service",
            service="open-vm-tools",
            enabled=True,
        )
    """

    service_operation: "PyinfraOperation"

    if host.get_fact(Which, command="systemctl"):
        service_operation = systemd.service

    elif host.get_fact(Which, command="rc-service"):
        service_operation = openrc.service

    elif host.get_fact(Which, command="initctl"):
        service_operation = upstart.service

    elif host.get_fact(Which, command="sv"):
        service_operation = runit.service

    elif (
        host.get_fact(Which, command="service")
        or host.get_fact(Link, path="/etc/init.d")
        or host.get_fact(Directory, path="/etc/init.d")
    ):
        service_operation = sysvinit.service

    # NOTE: important that we are not Linux here because /etc/rc.d will exist but checking it's
    # contents may trigger things (like a reboot: https://github.com/Fizzadar/pyinfra/issues/819)
    elif host.get_fact(Os) != "Linux" and bool(host.get_fact(Directory, path="/etc/rc.d")):
        service_operation = bsdinit.service

    else:
        raise OperationError(
            ("No init system found " "(no systemctl, initctl, /etc/init.d or /etc/rc.d found)"),
        )

    yield from service_operation._inner(
        service=service,
        running=running,
        restarted=restarted,
        reloaded=reloaded,
        command=command,
        enabled=enabled,
    )


@operation()
def packages(
    packages: str | list[str],
    present=True,
):
    """
    Add or remove system packages. This command checks for the presence of all the
    system package managers pyinfra can handle and executes the relevant operation.

    + packages: list of packages to ensure
    + present: whether the packages should be installed

    **Example:**

    .. code:: python

        server.packages(
            name="Install Vim and vimpager",
            packages=["vimpager", "vim"],
        )
    """

    package_operation: "PyinfraOperation"

    # TODO: improve this - use LinuxDistribution fact + mapping with fallback below?
    # Here to be preferred on openSUSE which also provides aptitude
    # See: https://github.com/Fizzadar/pyinfra/issues/799
    if host.get_fact(Which, command="zypper"):
        package_operation = zypper.packages

    elif host.get_fact(Which, command="apk"):
        package_operation = apk.packages

    elif host.get_fact(Which, command="apt"):
        package_operation = apt.packages

    elif host.get_fact(Which, command="brew"):
        package_operation = brew.packages

    elif host.get_fact(Which, command="dnf"):
        package_operation = dnf.packages

    elif host.get_fact(Which, command="pacman"):
        package_operation = pacman.packages

    elif host.get_fact(Which, command="xbps-install") or host.get_fact(Which, command="xbps"):
        package_operation = xbps.packages

    elif host.get_fact(Which, command="yum"):
        package_operation = yum.packages

    elif host.get_fact(Which, command="pkg") or host.get_fact(Which, command="pkg_add"):
        package_operation = pkg.packages

    else:
        raise OperationError(
            (
                "No system package manager found "
                "(no apk, apt, brew, dnf, pacman, pkg, xbps, yum or zypper found)"
            ),
        )

    yield from package_operation._inner(packages=packages, present=present)


@operation()
def crontab(
    command: str,
    present=True,
    user: str | None = None,
    cron_name: str | None = None,
    minute="*",
    hour="*",
    month="*",
    day_of_week="*",
    day_of_month="*",
    special_time: str | None = None,
    interpolate_variables=False,
):
    """
    Add/remove/update crontab entries.

    + command: the command for the cron
    + present: whether this cron command should exist
    + user: the user whose crontab to manage
    + cron_name: name the cronjob so future changes to the command will overwrite
    + minute: which minutes to execute the cron
    + hour: which hours to execute the cron
    + month: which months to execute the cron
    + day_of_week: which day of the week to execute the cron
    + day_of_month: which day of the month to execute the cron
    + special_time: cron "nickname" time (@reboot, @daily, etc), overrides others
    + interpolate_variables: whether to interpolate variables in ``command``

    Cron commands:
        Unless ``name`` is specified the command is used to identify crontab entries.
        This means commands must be unique within a given users crontab. If you require
        multiple identical commands, provide a different name argument for each.

    Special times:
        When provided, ``special_time`` will be used instead of any values passed in
        for ``minute``/``hour``/``month``/``day_of_week``/``day_of_month``.

    **Example:**

    .. code:: python

        # simple example for a crontab
        server.crontab(
            name="Backup /etc weekly",
            command="/bin/tar cf /tmp/etc_bup.tar /etc",
            name="backup_etc",
            day_of_week=0,
            hour=1,
            minute=0,
        )
    """

    def comma_sep(value):
        if isinstance(value, (list, tuple)):
            return ",".join("{0}".format(v) for v in value)
        return value

    minute = comma_sep(minute)
    hour = comma_sep(hour)
    month = comma_sep(month)
    day_of_week = comma_sep(day_of_week)
    day_of_month = comma_sep(day_of_month)

    crontab = host.get_fact(Crontab, user=user)
    name_comment = "# pyinfra-name={0}".format(cron_name)

    existing_crontab = crontab.get(command)
    existing_crontab_command = command
    existing_crontab_match = command

    if not existing_crontab and cron_name:  # find the crontab by name if provided
        for cmd, details in crontab.items():
            if not details["comments"]:
                continue
            if name_comment in details["comments"]:
                existing_crontab = details
                existing_crontab_match = cmd
                existing_crontab_command = cmd

    exists = existing_crontab is not None

    edit_commands: list[str | StringCommand] = []
    temp_filename = host.get_temp_filename()

    if special_time:
        new_crontab_line = "{0} {1}".format(special_time, command)
    else:
        new_crontab_line = "{minute} {hour} {day_of_month} {month} {day_of_week} {command}".format(
            minute=minute,
            hour=hour,
            day_of_month=day_of_month,
            month=month,
            day_of_week=day_of_week,
            command=command,
        )

    existing_crontab_match = ".*{0}.*".format(existing_crontab_match)

    # Don't want the cron and it does exist? Remove the line
    if not present and exists:
        edit_commands.append(
            sed_replace(
                temp_filename,
                existing_crontab_match,
                "",
                interpolate_variables=interpolate_variables,
            ),
        )

    # Want the cron but it doesn't exist? Append the line
    elif present and not exists:
        if cron_name:
            if crontab:  # append a blank line if cron entries already exist
                edit_commands.append("echo '' >> {0}".format(temp_filename))
            edit_commands.append(
                "echo {0} >> {1}".format(
                    shlex.quote(name_comment),
                    temp_filename,
                ),
            )

        edit_commands.append(
            "echo {0} >> {1}".format(
                shlex.quote(new_crontab_line),
                temp_filename,
            ),
        )

    # We have the cron and it exists, do it's details? If not, replace the line
    elif present and exists:
        assert existing_crontab is not None
        if any(
            (
                special_time != existing_crontab.get("special_time"),
                try_int(minute) != existing_crontab.get("minute"),
                try_int(hour) != existing_crontab.get("hour"),
                try_int(month) != existing_crontab.get("month"),
                try_int(day_of_week) != existing_crontab.get("day_of_week"),
                try_int(day_of_month) != existing_crontab.get("day_of_month"),
                existing_crontab_command != command,
            ),
        ):
            edit_commands.append(
                sed_replace(
                    temp_filename,
                    existing_crontab_match,
                    new_crontab_line,
                    interpolate_variables=interpolate_variables,
                ),
            )

    if edit_commands:
        crontab_args = []
        if user:
            crontab_args.append("-u {0}".format(user))

        # List the crontab into a temporary file if it exists
        if crontab:
            yield "crontab -l {0} > {1}".format(" ".join(crontab_args), temp_filename)

        # Now yield any edits
        for edit_command in edit_commands:
            yield edit_command

        # Finally, use the tempfile to write a new crontab
        yield "crontab {0} {1}".format(" ".join(crontab_args), temp_filename)
    else:
        host.noop(
            "crontab {0} {1}".format(
                command,
                "exists" if present else "does not exist",
            ),
        )


@operation()
def group(group: str, present=True, system=False, gid: int | str | None = None):
    """
    Add/remove system groups.

    + group: name of the group to ensure
    + present: whether the group should be present or not
    + system: whether to create a system group
    + gid: use a specific groupid number

    System users:
        System users don't exist on BSD, so the argument is ignored for BSD targets.

    **Examples:**

    .. code:: python

        server.group(
            name="Create docker group",
            group="docker",
        )

        # multiple groups
        for group in ["wheel", "lusers"]:
            server.group(
                name=f"Create the group {group}",
                group=group,
            )
    """

    groups = host.get_fact(Groups)
    os_type = host.get_fact(Os)
    is_present = group in groups

    # Group exists but we don't want them?
    if not present and is_present:
        if os_type == "FreeBSD":
            yield "pw groupdel -n {0}".format(group)
        else:
            yield "groupdel {0}".format(group)

    # Group doesn't exist and we want it?
    elif present and not is_present:
        args = []

        # BSD doesn't do system users
        if system and "BSD" not in host.get_fact(Os):
            args.append("-r")

        if os_type == "FreeBSD":
            args.append("-n {0}".format(group))
        else:
            args.append(group)

        if gid:
            if os_type == "FreeBSD":
                args.append("-g {0}".format(gid))
            else:
                args.append("--gid {0}".format(gid))

        # Groups are often added by other operations (package installs), so check
        # for the group at runtime before adding.
        group_add_command = "groupadd"
        if os_type == "FreeBSD":
            group_add_command = "pw groupadd"
        yield "{0} {1}".format(group_add_command, " ".join(args))


@operation()
def user_authorized_keys(
    user: str,
    public_keys: str | list[str],
    group: str | None = None,
    delete_keys=False,
    authorized_key_directory: str | None = None,
    authorized_key_filename: str | None = None,
):
    """
    Manage `authorized_keys` of system users.

    + user: name of the user to ensure
    + public_keys: list of public keys to attach to this user, ``home`` must be specified
    + group: the users primary group
    + delete_keys: whether to remove any keys not specified in ``public_keys``

    Public keys:
        These can be provided as strings containing the public key or as a path to
        a public key file which pyinfra will read.

    **Examples:**

    .. code:: python

        server.user_authorized_keys(
            name="Ensure user has a public key",
            user="kevin",
            public_keys=["ed25519..."],
        )
    """

    if not authorized_key_directory:
        home = host.get_fact(Home, user=user)
        authorized_key_directory = f"{home}/.ssh"

    if not authorized_key_filename:
        authorized_key_filename = "authorized_keys"

    if isinstance(public_keys, str):
        public_keys = [public_keys]

    def read_any_pub_key_file(key):
        try_path = key
        if state.cwd:
            try_path = path.join(state.cwd, key)

        if path.exists(try_path):
            with open(try_path, "r") as f:
                return [key.strip() for key in f.readlines()]

        return [key.strip()]

    public_keys = [key for key_or_file in public_keys for key in read_any_pub_key_file(key_or_file)]

    # Ensure .ssh directory
    # note that this always outputs commands unless the SSH user has access to the
    # authorized_keys file, ie the SSH user is the user defined in this function
    yield from files.directory._inner(
        path=authorized_key_directory,
        user=user,
        group=group or user,
        mode=700,
    )

    authorized_key_file = f"{authorized_key_directory}/{authorized_key_filename}"

    if delete_keys:
        # Create a whole new authorized_keys file
        keys_file = StringIO(
            "{0}\n".format(
                "\n".join(public_keys),
            ),
        )

        # And ensure it exists
        yield from files.put._inner(
            src=keys_file,
            dest=authorized_key_file,
            user=user,
            group=group or user,
            mode=600,
        )

    else:
        # Ensure authorized_keys exists
        yield from files.file._inner(
            path=authorized_key_file,
            user=user,
            group=group or user,
            mode=600,
        )

        # And every public key is present
        for key in public_keys:
            yield from files.line._inner(path=authorized_key_file, line=key, ensure_newline=True)


@operation()
def user(
    user: str,
    present=True,
    home: str | None = None,
    shell: str | None = None,
    group: str | None = None,
    groups: list[str] | None = None,
    public_keys: str | list[str] | None = None,
    delete_keys=False,
    ensure_home=True,
    create_home=False,
    system=False,
    uid: int | None = None,
    comment: str | None = None,
    add_deploy_dir=True,
    unique=True,
    password: str | None = None,
):
    """
    Add/remove/update system users & their ssh `authorized_keys`.

    + user: name of the user to ensure
    + present: whether this user should exist
    + home: the users home directory
    + shell: the users shell
    + group: the users primary group
    + groups: the users secondary groups
    + public_keys: list of public keys to attach to this user, ``home`` must be specified
    + delete_keys: whether to remove any keys not specified in ``public_keys``
    + ensure_home: whether to ensure the ``home`` directory exists
    + create_home: whether to new user create home directories from the system skeleton
    + system: whether to create a system account
    + uid: use a specific userid number
    + comment: the user GECOS comment
    + add_deploy_dir: any public_key filenames are relative to the deploy directory
    + unique: prevent creating users with duplicate UID
    + password: set the encrypted password for the user

    Home directory:
        When ``ensure_home`` or ``public_keys`` are provided, ``home`` defaults to
        ``/home/{name}``. When ``create_home`` is ``True`` any newly created users
        will be created with the ``-m`` flag to build a new home directory from the
        systems skeleton directory.

    Public keys:
        These can be provided as strings containing the public key or as a path to
        a public key file which pyinfra will read.

    **Examples:**

    .. code:: python

        server.user(
            name="Ensure user is removed",
            user="kevin",
            present=False,
        )

        server.user(
            name="Ensure myweb user exists",
            user="myweb",
            shell="/bin/bash",
        )

        # multiple users
        for user in ["kevin", "bob"]:
            server.user(
                name=f"Ensure user {user} is removed",
                user=user,
                present=False,
            )
    """

    users = host.get_fact(Users)
    existing_groups = host.get_fact(Groups)
    existing_user = users.get(user)
    os_type = host.get_fact(Os)
    if groups is None:
        groups = []

    if home is None:
        home = "/home/{0}".format(user)
        if existing_user:
            home = existing_user.get("home", home)

    # User not wanted?
    if not present:
        if existing_user:
            if os_type == "FreeBSD":
                yield "pw userdel -n {0}".format(user)
            else:
                yield "userdel {0}".format(user)
        return

    # User doesn't exist but we want them?
    if present and existing_user is None:
        # Fix the case where a group of the same name already exists, tell useradd to use this
        # group rather than failing trying to create it.
        if not group and user in existing_groups:
            group = user

        # Create the user w/home/shell
        args = []

        if home:
            args.append("-d {0}".format(home))

        if shell:
            args.append("-s {0}".format(shell))

        if group:
            args.append("-g {0}".format(group))

        if groups:
            args.append("-G {0}".format(",".join(groups)))

        if system and "BSD" not in host.get_fact(Os):
            args.append("-r")

        if uid:
            if os_type == "FreeBSD":
                args.append("-u {0}".format(uid))
            else:
                args.append("--uid {0}".format(uid))

        if comment:
            args.append("-c '{0}'".format(comment))

        if not unique:
            args.append("-o")

        if create_home:
            args.append("-m")
        else:
            args.append("-M")

        if password:
            args.append("-p '{0}'".format(password))

        # Users are often added by other operations (package installs), so check
        # for the user at runtime before adding.
        add_user_command = "useradd"
        if os_type == "FreeBSD":
            add_user_command = "pw useradd"
            yield "{0} -n {2} {1}".format(
                add_user_command,
                " ".join(args),
                user,
            )
        else:
            yield "{0} {1} {2}".format(
                add_user_command,
                " ".join(args),
                user,
            )

    # User exists and we want them, check home/shell/keys/password
    else:
        args = []

        # Check homedir
        if home and existing_user["home"] != home:
            args.append("-d {0}".format(home))

        # Check shell
        if shell and existing_user["shell"] != shell:
            args.append("-s {0}".format(shell))

        # Check primary group
        if group and existing_user["group"] != group:
            args.append("-g {0}".format(group))

        # Check secondary groups, if defined
        if groups and set(existing_user["groups"]) != set(groups):
            args.append("-G {0}".format(",".join(groups)))

        if comment and existing_user["comment"] != comment:
            args.append("-c '{0}'".format(comment))

        if password and existing_user["password"] != password:
            args.append("-p '{0}'".format(password))

        # Need to mod the user?
        if args:
            if os_type == "FreeBSD":
                yield "pw usermod -n {1} {0}".format(" ".join(args), user)
            else:
                yield "usermod {0} {1}".format(" ".join(args), user)
            if comment:
                existing_user["comment"] = comment
            if home:
                existing_user["home"] = home
            if shell:
                existing_user["shell"] = shell
            if group:
                existing_user["group"] = group
            if groups:
                existing_user["groups"] = groups
            if password:
                existing_user["password"] = password

    # Ensure home directory ownership
    if ensure_home and home:
        yield from files.directory._inner(
            path=home,
            user=user,
            group=group or user,
            # Don't fail if the home directory exists as a link
            _no_fail_on_link=True,
        )

    # Add SSH keys
    if public_keys is not None:
        yield from user_authorized_keys._inner(
            user=user,
            public_keys=public_keys,
            group=group,
            delete_keys=delete_keys,
            authorized_key_directory="{0}/.ssh".format(home),
            authorized_key_filename=None,
        )


@operation()
def locale(
    locale: str,
    present=True,
):
    """
    Enable/Disable locale.

    + locale: name of the locale to enable/disable
    + present: whether this locale should be present or not

    **Examples:**

    .. code:: python

        server.locale(
            name="Ensure en_GB.UTF-8 locale is not present",
            locale="en_GB.UTF-8",
            present=False,
        )

        server.locale(
            name="Ensure en_GB.UTF-8 locale is present",
            locale="en_GB.UTF-8",
        )

    """

    locales = host.get_fact(Locales)

    logger.debug("Enabled locales: {0}".format(locales))

    locales_definitions_file = "/etc/locale.gen"

    # Find the matching line in /etc/locale.gen
    matching_lines = host.get_fact(
        FindInFile, path=locales_definitions_file, pattern=rf"^.*{locale}[[:space:]]\+.*$"
    )

    if not matching_lines:
        raise OperationError(f"Locale {locale} not found in {locales_definitions_file}")

    if len(matching_lines) > 1:
        raise OperationError(f"Multiple locales matches for {locale} in {locales_definitions_file}")

    matching_line = matching_lines[0]

    # Remove locale
    if not present and locale in locales:
        logger.debug(f"Removing locale {locale}")

        yield from files.line._inner(
            path=locales_definitions_file, line=f"^{matching_line}$", replace=f"# {matching_line}"
        )

        yield "locale-gen"

    # Add locale
    if present and locale not in locales:
        logger.debug(f"Adding locale {locale}")

        yield from files.replace._inner(
            path=locales_definitions_file,
            text=f"^{matching_line}$",
            replace=f"{matching_line}".replace("# ", ""),
        )

        yield "locale-gen"


@operation()
def security_limit(
    domain: str,
    limit_type: str,
    item: str,
    value: int,
):
    """
    Edit /etc/security/limits.conf configuration.

    + domain: the domain (user, group, or wildcard) for the limit
    + limit_type: the type of limit (hard or soft)
    + item: the item to limit (e.g., nofile, nproc)
    + value: the value for the limit

    **Example:**

    .. code:: python

        security_limit(
            name="Set nofile limit for all users",
            domain='*',
            limit_type='soft',
            item='nofile',
            value=1024,
        )
    """

    line_format = f"{domain}\t{limit_type}\t{item}\t{value}"

    yield from files.line._inner(
        path="/etc/security/limits.conf",
        line=f"^{domain}[[:space:]]+{limit_type}[[:space:]]+{item}",
        replace=line_format,
    )



try:
    import importlib_metadata
except ImportError:
    import importlib.metadata as importlib_metadata  # type: ignore[no-redef]
from os import path
from typing import Iterable, Optional, Set

from packaging.markers import Marker
from packaging.requirements import Requirement
from packaging.specifiers import SpecifierSet
from packaging.version import Version

from pyinfra import __version__, state

from .exceptions import PyinfraError


class ConfigDefaults:
    # % of hosts which have to fail for all operations to stop
    FAIL_PERCENT: Optional[int] = None
    # Seconds to timeout SSH connections
    CONNECT_TIMEOUT: int = 10
    # Temporary directory (on the remote side) to use for caching any files/downloads, the default
    # None value first tries to load the hosts' temporary directory configured via "TMPDIR" env
    # variable, falling back to DEFAULT_TEMP_DIR if not set.
    TEMP_DIR: Optional[str] = None
    DEFAULT_TEMP_DIR: str = "/tmp"
    # Gevent pool size (defaults to #of target hosts)
    PARALLEL: int = 0
    # Specify the required pyinfra version (using PEP 440 setuptools specifier)
    REQUIRE_PYINFRA_VERSION: Optional[str] = None
    # Specify any required packages (either using PEP 440 or a requirements file)
    # Note: this can also include pyinfra potentially replacing REQUIRE_PYINFRA_VERSION
    REQUIRE_PACKAGES: Optional[str] = None
    # All these can be overridden inside individual operation calls:
    # Switch to this user (from ssh_user) using su before executing operations
    SU_USER: Optional[str] = None
    USE_SU_LOGIN: bool = False
    SU_SHELL: bool = False
    PRESERVE_SU_ENV: bool = False
    # Use sudo and optional user
    SUDO: bool = False
    SUDO_USER: Optional[str] = None
    PRESERVE_SUDO_ENV: bool = False
    USE_SUDO_LOGIN: bool = False
    SUDO_PASSWORD: Optional[str] = None
    # Use doas and optional user
    DOAS: bool = False
    DOAS_USER: Optional[str] = None
    # Only show errors but don't count as failure
    IGNORE_ERRORS: bool = False
    # Shell to use to execute commands
    SHELL: str = "sh"


config_defaults = {key: value for key, value in ConfigDefaults.__dict__.items() if key.isupper()}


def check_pyinfra_version(version: str):
    if not version:
        return
    running_version = Version(__version__)
    required_versions = SpecifierSet(version)

    if running_version not in required_versions:
        raise PyinfraError(
            f"pyinfra version requirement not met (requires {version}, running {__version__})"
        )


def _check_requirements(requirements: Iterable[str]) -> Set[Requirement]:
    """
    Check whether each of the given requirements and all their dependencies are
    installed.

    Or more precisely, this checks that each of the given *requirements* is
    satisfied by some installed *distribution package*, and so on recursively
    for each of the dependencies of those distribution packages. The terminology
    here is as follows:

    * A *distribution package* is essentially a thing that can be installed with
      ``pip``, from an sdist or wheel or Git repo or so on.
    * A *requirement* is the expectation that a distribution package satisfying
      some constraint is installed.
    * A *dependency* is a requirement specified by a distribution package (as
      opposed to the requirements passed in to this function).

    So what this function does is start from the given requirements, for each
    one check that it is satisfied by some installed distribution package, and
    if so recursively perform the same check on all the dependencies of that
    distribution package. In short, it's traversing the graph of package
    requirements. It stops whenever it finds a requirement that is not satisfied
    (i.e. a required package that is not installed), or when it runs out of
    requirements to check.

    .. note::
        This is basically equivalent to ``pkg_resources.require()`` except that
        when ``require()`` succeeds, it will return the list of distribution
        packages that satisfy the given requirements and their dependencies, and
        when it fails, it will raise an exception. This function just returns
        the requirements which were not satisfied instead.

    :param requirements: The requirements to check for in the set of installed
        packages (along with their dependencies).
    :return: The set of requirements that were not satisfied, which will be
        an empty set if all requirements (recursively) were satisfied.
    """

    # Based on pkg_resources.require() from setuptools. The implementation of
    # hbutils.system.check_reqs() from the hbutils package was also helpful in
    # clarifying what this is supposed to do.

    reqs_to_check: Set[Requirement] = set(Requirement(r) for r in requirements)
    reqs_satisfied: Set[Requirement] = set()
    reqs_not_satisfied: Set[Requirement] = set()

    while reqs_to_check:
        req = reqs_to_check.pop()
        assert req not in reqs_satisfied and req not in reqs_not_satisfied

        # Check for an installed distribution package with the right name and version
        try:
            dist = importlib_metadata.distribution(req.name)
        except importlib_metadata.PackageNotFoundError:
            # No installed package with the right name
            # This would raise a DistributionNotFound error from pkg_resources.require()
            reqs_not_satisfied.add(req)
            continue

        if dist.version not in req.specifier:
            # There is a distribution with the right name but wrong version
            # This would raise a VersionConflict error from pkg_resources.require()
            reqs_not_satisfied.add(req)
            continue

        reqs_satisfied.add(req)

        # If the distribution package has dependencies of its own, go through
        # those dependencies and for each one add it to the set to be checked if
        # - it's unconditional (no marker)
        # - or it's conditional and the condition is satisfied (the marker
        #   evaluates to true) in the current environment
        # Markers can check things like the Python version and system version
        # etc., and/or they can check which extras of the distribution package
        # were required. To facilitate checking extras we have to pass the extra
        # in the environment when calling Marker.evaluate().
        if dist.requires:
            if req.extras:
                extras_envs = [{"extra": extra} for extra in req.extras]

                def evaluate_marker(marker: Marker) -> bool:
                    return any(map(marker.evaluate, extras_envs))

            else:

                def evaluate_marker(marker: Marker) -> bool:
                    return marker.evaluate()

            for dist_req_str in dist.requires:
                dist_req = Requirement(dist_req_str)
                if dist_req in reqs_satisfied or dist_req in reqs_not_satisfied:
                    continue
                if (not dist_req.marker) or evaluate_marker(dist_req.marker):
                    reqs_to_check.add(dist_req)

    return reqs_not_satisfied


def check_require_packages(requirements_config):
    if not requirements_config:
        return

    if isinstance(requirements_config, (list, tuple)):
        requirements = requirements_config
    else:
        with open(path.join(state.cwd or "", requirements_config), encoding="utf-8") as f:
            requirements = [line.split("#egg=")[-1] for line in f.read().splitlines()]

    requirements_not_met = _check_requirements(requirements)
    if requirements_not_met:
        raise PyinfraError(
            "Deploy requirements ({0}) not met: missing {1}".format(
                requirements_config, ", ".join(str(r) for r in requirements_not_met)
            )
        )


config_checkers = {
    "REQUIRE_PYINFRA_VERSION": check_pyinfra_version,
    "REQUIRE_PACKAGES": check_require_packages,
}


class Config(ConfigDefaults):
    """
    The default/base configuration options for a pyinfra deploy.
    """

    def __init__(self, **kwargs):
        # Always apply some env
        env = kwargs.pop("ENV", {})
        self.ENV = env

        config = config_defaults.copy()
        config.update(kwargs)

        for key, value in config.items():
            setattr(self, key, value)

    def __setattr__(self, key, value):
        super().__setattr__(key, value)

        checker = config_checkers.get(key)
        if checker:
            checker(value)

    def get_current_state(self):
        return [(key, getattr(self, key)) for key in config_defaults.keys()]

    def set_current_state(self, config_state):
        for key, value in config_state:
            setattr(self, key, value)

    def lock_current_state(self):
        self._locked_config = self.get_current_state()

    def reset_locked_state(self):
        self.set_current_state(self._locked_config)

    def copy(self) -> "Config":
        return Config(**dict(self.get_current_state()))



from os import path

from pyinfra import inventory
from pyinfra.context import ctx_inventory, ctx_state

from ..paramiko_util import PatchSSHTestCase
from .util import run_cli


class TestCliInventory(PatchSSHTestCase):
    def test_load_deploy_group_data(self):
        ctx_state.reset()
        ctx_inventory.reset()

        hosts = ["somehost", "anotherhost", "someotherhost"]
        result = run_cli(
            "-y",
            ",".join(hosts),
            path.join("tests", "test_cli", "deploy", "deploy.py"),
            f'--chdir={path.join("tests", "test_cli", "deploy")}',
        )
        assert result.exit_code == 0, result.stdout

        assert inventory.data.get("hello") == "world"
        assert "leftover_data" in inventory.group_data
        assert inventory.group_data["leftover_data"].get("still_parsed") == "never_used"
        assert inventory.group_data["leftover_data"].get("_global_arg") == "gets_parsed"

    def test_load_group_data(self):
        ctx_state.reset()
        ctx_inventory.reset()

        hosts = ["somehost", "anotherhost", "someotherhost"]
        result = run_cli(
            "-y",
            ",".join(hosts),
            f'--group-data={path.join("tests", "test_cli", "deploy", "group_data")}',
            "exec",
            "uptime",
        )
        assert result.exit_code == 0, result.stdout

        assert inventory.data.get("hello") == "world"
        assert "leftover_data" in inventory.group_data
        assert inventory.group_data["leftover_data"].get("still_parsed") == "never_used"
        assert inventory.group_data["leftover_data"].get("_global_arg") == "gets_parsed"

    def test_load_group_data_file(self):
        ctx_state.reset()
        ctx_inventory.reset()

        hosts = ["somehost", "anotherhost", "someotherhost"]
        filename = path.join("tests", "test_cli", "deploy", "group_data", "leftover_data.py")
        result = run_cli(
            "-y",
            ",".join(hosts),
            f"--group-data={filename}",
            "exec",
            "uptime",
        )
        assert result.exit_code == 0, result.stdout

        assert "hello" not in inventory.data
        assert "leftover_data" in inventory.group_data
        assert inventory.group_data["leftover_data"].get("still_parsed") == "never_used"
        assert inventory.group_data["leftover_data"].get("_global_arg") == "gets_parsed"



import logging
import sys
import warnings
from fnmatch import fnmatch
from os import chdir as os_chdir, getcwd, path
from typing import Iterable, List, Tuple, Union

import click

from pyinfra import __version__, logger, state
from pyinfra.api import Config, State
from pyinfra.api.connect import connect_all, disconnect_all
from pyinfra.api.exceptions import NoGroupError, PyinfraError
from pyinfra.api.facts import get_facts
from pyinfra.api.operations import run_ops
from pyinfra.api.state import StateStage
from pyinfra.api.util import get_kwargs_str
from pyinfra.context import ctx_config, ctx_inventory, ctx_state
from pyinfra.operations import server

from .commands import get_facts_and_args, get_func_and_args
from .exceptions import CliError, UnexpectedExternalError, UnexpectedInternalError, WrappedError
from .inventory import make_inventory
from .log import setup_logging
from .prints import (
    print_facts,
    print_inventory,
    print_meta,
    print_results,
    print_state_operations,
    print_support_info,
)
from .util import exec_file, load_deploy_file, load_func, parse_cli_arg
from .virtualenv import init_virtualenv


def _exit():
    if ctx_state.isset() and state.failed_hosts:
        sys.exit(1)
    sys.exit(0)


def _print_support(ctx, param, value):
    if not value:
        return

    logger.info("--> Support information:")
    print_support_info()
    ctx.exit()


@click.command()
@click.argument("inventory", nargs=1, type=click.Path(exists=False))
@click.argument("operations", nargs=-1, required=True, type=click.Path(exists=False))
@click.option(
    "verbosity",
    "-v",
    count=True,
    help="Print meta (-v), input (-vv) and output (-vvv).",
)
@click.option(
    "--dry",
    is_flag=True,
    default=False,
    help="Don't execute operations on the target hosts.",
)
@click.option(
    "-y",
    "--yes",
    is_flag=True,
    default=False,
    help="Execute operations immediately on hosts without prompt or checking for changes.",
    envvar="PYINFRA_YES",
    show_envvar=True,
)
@click.option(
    "--limit",
    help="Restrict the target hosts by name and group name.",
    multiple=True,
)
@click.option("--fail-percent", type=int, help="% of hosts that need to fail before exiting early.")
@click.option(
    "--data",
    multiple=True,
    help="Override data values, format key=value.",
)
@click.option(
    "--group-data",
    multiple=True,
    help="Paths to load additional group data from (overrides matching keys).",
)
@click.option(
    "--config",
    "config_filename",
    help="Specify config file to use (default: config.py).",
    default="config.py",
)
@click.option(
    "--chdir",
    help="Set the working directory before executing.",
)
# Auth args
@click.option(
    "--sudo",
    is_flag=True,
    default=False,
    help="Whether to execute operations with sudo.",
)
@click.option("--sudo-user", help="Which user to sudo when sudoing.")
@click.option(
    "--use-sudo-password",
    is_flag=True,
    default=False,
    help="Whether to use a password with sudo.",
)
@click.option("--su-user", help="Which user to su to.")
@click.option("--shell-executable", help='Shell to use (ex: "sh", "cmd", "ps").')
# Operation flow args
@click.option("--parallel", type=int, help="Number of operations to run in parallel.")
@click.option(
    "--no-wait",
    is_flag=True,
    default=False,
    help="Don't wait between operations for hosts.",
)
@click.option(
    "--serial",
    is_flag=True,
    default=False,
    help="Run operations in serial, host by host.",
)
# SSH connector args
# TODO: remove the non-ssh-prefixed variants
@click.option("--ssh-user", "--user", "ssh_user", help="SSH user to connect as.")
@click.option("--ssh-port", "--port", "ssh_port", type=int, help="SSH port to connect to.")
@click.option("--ssh-key", "--key", "ssh_key", type=click.Path(), help="SSH Private key filename.")
@click.option(
    "--ssh-key-password",
    "--key-password",
    "ssh_key_password",
    help="SSH Private key password.",
)
@click.option("--ssh-password", "--password", "ssh_password", help="SSH password.")
# Eager commands (pyinfra --support)
@click.option(
    "--support",
    is_flag=True,
    is_eager=True,
    callback=_print_support,
    help="Print useful information for support and exit.",
)
# Debug args
@click.option(
    "--debug",
    is_flag=True,
    default=False,
    help="Print debug logs from pyinfra.",
)
@click.option(
    "--debug-all",
    is_flag=True,
    default=False,
    help="Print debug logs from all packages including pyinfra.",
)
@click.option(
    "--debug-facts",
    is_flag=True,
    default=False,
    help="Print facts after generating operations and exit.",
)
@click.option(
    "--debug-operations",
    is_flag=True,
    default=False,
    help="Print operations after generating and exit.",
)
@click.version_option(
    version=__version__,
    prog_name="pyinfra",
    message="%(prog)s: v%(version)s",
)
def cli(*args, **kwargs):
    """
    pyinfra manages the state of one or more servers. It can be used for
    app/service deployment, config management and ad-hoc command execution.

    Documentation: docs.pyinfra.com

    # INVENTORY

    \b
    + a file (inventory.py)
    + hostname (host.net)
    + Comma separated hostnames:
      host-1.net,host-2.net,@local

    # OPERATIONS

    \b
    # Run one or more deploys against the inventory
    pyinfra INVENTORY deploy_web.py [deploy_db.py]...

    \b
    # Run a single operation against the inventory
    pyinfra INVENTORY server.user pyinfra home=/home/pyinfra

    \b
    # Execute an arbitrary command against the inventory
    pyinfra INVENTORY exec -- echo "hello world"

    \b
    # Run one or more facts against the inventory
    pyinfra INVENTORY fact server.LinuxName [server.Users]...
    pyinfra INVENTORY fact files.File path=/path/to/file...

    \b
    # Debug the inventory hosts and data
    pyinfra INVENTORY debug-inventory
    """

    try:
        _main(*args, **kwargs)
    except (CliError, UnexpectedExternalError):
        raise
    except PyinfraError as e:
        # Re-raise "expected" pyinfra exceptions with our click exception wrapper
        raise WrappedError(e)
    except Exception as e:
        # Re-raise any unexpected internal exceptions as UnexpectedInternalError
        raise UnexpectedInternalError(e)
    finally:
        if ctx_state.isset() and state.initialised:
            logger.info("--> Disconnecting from hosts...")
            # Triggers any executor disconnect requirements
            disconnect_all(state)


class CliCommands:
    DEBUG_INVENTORY = "DEBUG_INVENTORY"
    FACT = "FACT"
    SHELL = "SHELL"
    DEPLOY_FILES = "DEPLOY_FILES"
    FUNC = "FUNC"


def _main(
    inventory,
    operations: Union[List, Tuple],
    verbosity: int,
    chdir: str,
    ssh_user,
    ssh_port: int,
    ssh_key,
    ssh_key_password: str,
    ssh_password: str,
    shell_executable,
    sudo: bool,
    sudo_user: str,
    use_sudo_password: bool,
    su_user: str,
    parallel: int,
    fail_percent: int,
    data,
    group_data,
    config_filename: str,
    dry: bool,
    yes: bool,
    limit: Iterable,
    no_wait: bool,
    serial: bool,
    debug: bool,
    debug_all: bool,
    debug_facts: bool,
    debug_operations: bool,
    support: bool = False,
):
    # Setup working directory
    #
    if chdir:
        os_chdir(chdir)

    # Setup logging & Bootstrap/Venv
    #
    _setup_log_level(debug, debug_all)
    init_virtualenv()

    # Check operations are valid and setup commands
    #
    original_operations, operations, command, chdir = _validate_operations(operations, chdir)

    # Setup state, config & inventory
    #
    state = _setup_state(verbosity, yes)
    config = Config()
    ctx_config.set(config)

    # Update Config & Override Data
    #
    config = _set_config(
        config,
        config_filename,
        sudo,
        sudo_user,
        use_sudo_password,
        su_user,
        parallel,
        shell_executable,
        fail_percent,
        yes,
    )
    override_data = _set_override_data(
        data,
        ssh_user,
        ssh_key,
        ssh_key_password,
        ssh_port,
        ssh_password,
    )

    if yes is False:
        _set_fail_prompts(state, config)

    # Load up the inventory from the filesystem
    #
    logger.info("--> Loading inventory...")
    inventory = make_inventory(
        inventory,
        cwd=state.cwd,
        override_data=override_data,
        group_data_directories=group_data,
    )
    ctx_inventory.set(inventory)

    # Now that we have inventory, apply --limit config override
    initial_limit = _apply_inventory_limit(inventory, limit)

    # Initialise the state
    state.init(inventory, config, initial_limit=initial_limit)

    if command == CliCommands.DEBUG_INVENTORY:
        print_inventory(state)
        _exit()

    # Connect to the hosts & start handling the user commands
    #
    logger.info("--> Connecting to hosts...")
    state.set_stage(StateStage.Connect)
    connect_all(state)

    logger.info("--> Preparing operations...")
    state.set_stage(StateStage.Prepare)
    can_diff, state, config = _handle_commands(
        state, config, command, original_operations, operations
    )

    # Print proposed changes, execute unless --dry, and exit
    #
    if can_diff:
        if yes:
            logger.info("--> Skipping change detection")
        else:
            logger.info("--> Detected changes:")
            print_meta(state)
            click.echo(
                """
    Detected changes may not include every change pyinfra will execute.
    Hidden side effects of operations may alter behaviour of future operations,
    this will be shown in the results. The remote state will always be updated
    to reflect the state defined by the input operations.""",
                err=True,
            )

    # If --debug-facts or --debug-operations, print and exit
    if debug_facts or debug_operations:
        if debug_operations:
            print_state_operations(state)

        _exit()

    if dry:
        _exit()

    if (
        can_diff
        and not yes
        and not _do_confirm("Detected changes displayed above, skip this step with -y")
    ):
        _exit()

    logger.info("--> Beginning operation run...")
    state.set_stage(StateStage.Execute)
    run_ops(state, serial=serial, no_wait=no_wait)

    logger.info("--> Results:")
    state.set_stage(StateStage.Disconnect)
    print_results(state)
    _exit()


def _do_confirm(msg: str) -> bool:
    click.echo(err=True)
    click.echo(f"    {msg}", err=True)
    warning_count = state.get_warning_counter()
    if warning_count > 0:
        click.secho(
            f"    {warning_count} warnings shown during change detection, see above",
            fg="yellow",
            err=True,
        )
    confirm_msg = "    Press enter to execute..."
    click.echo(confirm_msg, err=True, nl=False)
    v = input()
    if v:
        click.echo(f"    Unexpected user input: {v}", err=True)
        return False
    # Go up, clear the line, go up again - as if the confirmation statement was never here!
    click.echo(
        "\033[1A{0}\033[1A".format("".join(" " for _ in range(len(confirm_msg)))),
        err=True,
        nl=False,
    )
    click.echo(err=True)
    return True


# Setup
#
def _setup_log_level(debug, debug_all):
    if not debug and not sys.warnoptions:
        warnings.simplefilter("ignore")

    log_level = logging.INFO
    if debug or debug_all:
        log_level = logging.DEBUG

    other_log_level = None
    if debug_all:
        other_log_level = logging.DEBUG

    setup_logging(log_level, other_log_level)


def _validate_operations(operations, chdir):
    # Make a copy before we overwrite
    original_operations = operations

    # Debug (print) inventory + group data
    if operations[0] == "debug-inventory":
        command = CliCommands.DEBUG_INVENTORY

    # Get one or more facts
    elif operations[0] == "fact":
        command = CliCommands.FACT
        operations = get_facts_and_args(operations[1:])

    # Execute a raw command with server.shell
    elif operations[0] == "exec":
        command = CliCommands.SHELL
        operations = operations[1:]

    # Execute one or more deploy files
    elif all(cmd.endswith(".py") for cmd in operations):
        command = CliCommands.DEPLOY_FILES

        filenames = []

        for filename in operations[0:]:
            if path.exists(filename):
                filenames.append(filename)
                continue
            if chdir and filename.startswith(chdir):
                correct_filename = path.relpath(filename, chdir)
                logger.warning(
                    (
                        "Fixing deploy filename under `--chdir` argument: "
                        f"{filename} -> {correct_filename}"
                    ),
                )
                filenames.append(correct_filename)
                continue
            raise CliError(
                "No deploy file: {0}".format(
                    path.join(chdir, filename) if chdir else filename,
                ),
            )

        operations = filenames

    # Load a function (op or deploy) directly with arguments
    elif "." in operations[0]:
        command = CliCommands.FUNC
        operations = get_func_and_args(operations)

    else:
        raise CliError(
            """Invalid operations: {0}

    Operation usage:
    pyinfra INVENTORY deploy_web.py [deploy_db.py]...
    pyinfra INVENTORY server.user pyinfra home=/home/pyinfra
    pyinfra INVENTORY exec -- echo "hello world"
    pyinfra INVENTORY fact os [users]...""".format(
                operations,
            ),
        )

    return original_operations, operations, command, chdir


def _set_verbosity(state, verbosity):
    if verbosity > 0:
        state.print_fact_info = True
        state.print_noop_info = True

    if verbosity > 1:
        state.print_input = state.print_fact_input = True

    if verbosity > 2:
        state.print_output = state.print_fact_output = True

    return state


def _setup_state(verbosity, yes):
    cwd = getcwd()
    if cwd not in sys.path:  # ensure cwd is present in sys.path
        sys.path.append(cwd)

    state = State(check_for_changes=not yes)
    state.cwd = cwd
    ctx_state.set(state)

    state = _set_verbosity(state, verbosity)
    return state


def _set_config(
    config,
    config_filename,
    sudo,
    sudo_user,
    use_sudo_password,
    su_user,
    parallel,
    shell_executable,
    fail_percent,
    yes,
):
    logger.info("--> Loading config...")

    # Load up any config.py from the filesystem
    if state.cwd:
        config_filename = path.join(state.cwd, config_filename)
    if path.exists(config_filename):
        exec_file(config_filename)

    # Lock the current config, this allows us to restore this version after
    # executing deploy files that may alter them.
    config.lock_current_state()

    # Arg based config overrides
    if sudo:
        config.SUDO = True
        if sudo_user:
            config.SUDO_USER = sudo_user

    if use_sudo_password:
        config.USE_SUDO_PASSWORD = use_sudo_password

    if su_user:
        config.SU_USER = su_user

    if parallel:
        config.PARALLEL = parallel

    if shell_executable:
        config.SHELL = None if shell_executable in ("None", "null") else shell_executable

    if fail_percent is not None:
        config.FAIL_PERCENT = fail_percent

    return config


def _set_override_data(
    data,
    ssh_user,
    ssh_key,
    ssh_key_password,
    ssh_port,
    ssh_password,
):
    override_data = {}

    for arg in data:
        key, value = arg.split("=", 1)
        override_data[key] = value

    override_data = {key: parse_cli_arg(value) for key, value in override_data.items()}

    for key, value in (
        ("ssh_user", ssh_user),
        ("ssh_key", ssh_key),
        ("ssh_key_password", ssh_key_password),
        ("ssh_port", ssh_port),
        ("ssh_password", ssh_password),
    ):
        if value:
            override_data[key] = value

    return override_data


def _set_fail_prompts(state: State, config: Config) -> None:
    # Set fail percent to zero, meaning we'll raise an exception for any fail,
    # and we can capture + prompt the user to continue/exit.
    config.FAIL_PERCENT = 0

    def should_raise_failed_hosts(state: State) -> bool:
        if state.current_stage == StateStage.Connect:
            return not _do_confirm("One of more hosts failed to connect, continue?")
        return not _do_confirm("One of more hosts failed, continue?")

    state.should_raise_failed_hosts = should_raise_failed_hosts


def _apply_inventory_limit(inventory, limit):
    initial_limit = None
    if limit:
        all_limit_hosts = []

        for limiter in limit:
            try:
                limit_hosts = inventory.get_group(limiter)
            except NoGroupError:
                limit_hosts = [host for host in inventory if fnmatch(host.name, limiter)]

            if not limit_hosts:
                logger.warning("No host matches found for --limit pattern: {0}".format(limiter))

            all_limit_hosts.extend(limit_hosts)
        initial_limit = list(set(all_limit_hosts))

    return initial_limit


# Operations Execution
#
def _handle_commands(state, config, command, original_operations, operations):
    if command is CliCommands.FACT:
        state, fact_data = _run_fact_operations(state, config, operations)
        print_facts(fact_data)
        _exit()

    can_diff = True

    if command == CliCommands.SHELL:
        state = _prepare_exec_operations(state, config, operations)
        can_diff = False

    elif command == CliCommands.DEPLOY_FILES:
        state, config, operations = _prepare_deploy_operations(state, config, operations)

    elif command == CliCommands.FUNC:
        state, kwargs = _prepare_func_operations(
            state,
            config,
            operations,
            original_operations,
        )

    return can_diff, state, config


def _run_fact_operations(state, config, operations):
    logger.info("--> Gathering facts...")

    state.print_fact_info = True
    fact_data = {}

    for i, command in enumerate(operations):
        fact_cls, args, kwargs = command
        fact_key = fact_cls.name

        if args or kwargs:
            _fact_args = args or ""
            _fact_details = " ({0})".format(get_kwargs_str(kwargs)) if kwargs else ""
            fact_key = "{0}{1}{2}".format(fact_cls.name, _fact_args, _fact_details)

        try:
            fact_data[fact_key] = get_facts(
                state,
                fact_cls,
                args=args,
                kwargs=kwargs,
                apply_failed_hosts=False,
            )
        except PyinfraError:
            pass

    return state, fact_data


def _prepare_exec_operations(state, config, operations):
    state.print_output = True
    load_func(
        state,
        server.shell,
        " ".join(operations),
    )
    return state


def _prepare_deploy_operations(state, config, operations):
    logger.info("--> Preparing Operations...")

    # Number of "steps" to make = number of files * number of hosts
    for i, filename in enumerate(operations):
        _log_styled_msg = click.style(filename, bold=True)
        logger.info("Loading: {0}".format(_log_styled_msg))

        state.current_op_file_number = i
        load_deploy_file(state, filename)

        # Remove any config changes introduced by the deploy file & any includes
        config.reset_locked_state()

    return state, config, operations


def _prepare_func_operations(state, config, operations, original_operations):
    logger.info("--> Preparing operation...")

    op, args = operations
    args, kwargs = args

    load_func(state, op, *args, **kwargs)

    return state, kwargs



"""
Manage ZFS filesystems.
"""

from pyinfra.api import FactBase, ShortFactBase


def _process_zfs_props_table(output):
    datasets: dict = {}
    for line in output:
        dataset, property, value, source = tuple(line.split("\t"))
        if dataset not in datasets:
            datasets[dataset] = {}
        datasets[dataset][property] = value
    return datasets


class Pools(FactBase):
    def command(self):
        return "zpool get -H all"

    @staticmethod
    def process(output):
        return _process_zfs_props_table(output)


class Datasets(FactBase):
    def command(self):
        return "zfs get -H all"

    @staticmethod
    def process(output):
        return _process_zfs_props_table(output)


class Filesystems(ShortFactBase):
    fact = Datasets

    @staticmethod
    def process_data(data):
        return {name: props for name, props in data.items() if props.get("type") == "filesystem"}


class Snapshots(ShortFactBase):
    fact = Datasets

    @staticmethod
    def process_data(data):
        return {name: props for name, props in data.items() if props.get("type") == "snapshot"}


class Volumes(ShortFactBase):
    fact = Datasets

    @staticmethod
    def process_data(data):
        return {name: props for name, props in data.items() if props.get("type") == "volume"}



from __future__ import annotations

import json

from pyinfra.api import FactBase


class LxdContainers(FactBase):
    """
    Returns a list of running LXD containers
    """

    def command(self) -> str:
        return "lxc list --format json --fast"

    def requires_command(self) -> str:
        return "lxc"

    default = list

    def process(self, output):
        output = list(output)
        assert len(output) == 1
        return json.loads(output[0])



"""
Manage brew packages on mac/OSX. See https://brew.sh/
"""

from __future__ import annotations

import urllib.parse

from pyinfra import host
from pyinfra.api import operation
from pyinfra.facts.brew import BrewCasks, BrewPackages, BrewTaps, BrewVersion, new_cask_cli

from .util.packaging import ensure_packages


@operation(is_idempotent=False)
def update():
    """
    Updates brew repositories.
    """

    yield "brew update"


_update = update  # noqa: E305


@operation(is_idempotent=False)
def upgrade():
    """
    Upgrades all brew packages.
    """

    yield "brew upgrade"


_upgrade = upgrade  # noqa: E305


@operation()
def packages(
    packages: str | list[str] | None = None,
    present=True,
    latest=False,
    update=False,
    upgrade=False,
):
    """
    Add/remove/update brew packages.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + latest: whether to upgrade packages without a specified version
    + update: run ``brew update`` before installing packages
    + upgrade: run ``brew upgrade`` before installing packages

    Versions:
        Package versions can be pinned like brew: ``<pkg>@<version>``.

    **Examples:**

    .. code:: python

        # Update package list and install packages
        brew.packages(
            name='Install Vim and vimpager',
            packages=["vimpager", "vim"],
            update=True,
        )

        # Install the latest versions of packages (always check)
        brew.packages(
            name="Install latest Vim",
            packages=["vim"],
            latest=True,
        )
    """

    if update:
        yield from _update._inner()

    if upgrade:
        yield from _upgrade._inner()

    yield from ensure_packages(
        host,
        packages,
        host.get_fact(BrewPackages),
        present,
        install_command="brew install",
        uninstall_command="brew uninstall",
        upgrade_command="brew upgrade",
        version_join="@",
        latest=latest,
    )


def cask_args():
    return ("", " --cask") if new_cask_cli(host.get_fact(BrewVersion)) else ("cask ", "")


@operation(is_idempotent=False)
def cask_upgrade():
    """
    Upgrades all brew casks.
    """

    yield "brew %supgrade%s" % cask_args()


@operation()
def casks(
    casks: str | list[str] | None = None,
    present=True,
    latest=False,
    upgrade=False,
):
    """
    Add/remove/update brew casks.

    + casks: list of casks to ensure
    + present: whether the casks should be installed
    + latest: whether to upgrade casks without a specified version
    + upgrade: run brew cask upgrade before installing casks

    Versions:
        Cask versions can be pinned like brew: ``<pkg>@<version>``.

    **Example:**

    .. code:: python

        brew.casks(
            name='Upgrade and install the latest cask',
            casks=["godot"],
            upgrade=True,
            latest=True,
        )

    """

    if upgrade:
        yield from cask_upgrade._inner()

    args = cask_args()

    yield from ensure_packages(
        host,
        casks,
        host.get_fact(BrewCasks),
        present,
        install_command="brew %sinstall%s" % args,
        uninstall_command="brew %suninstall%s" % args,
        upgrade_command="brew %supgrade%s" % args,
        version_join="@",
        latest=latest,
    )


@operation()
def tap(src: str | None = None, present=True, url: str | None = None):
    """
    Add/remove brew taps.

    + src: the name of the tap
    + present: whether this tap should be present or not
    + url: the url of the tap. See https://docs.brew.sh/Taps

    **Examples:**

    .. code:: python

        brew.tap(
            name="Add a brew tap",
            src="includeos/includeos",
        )

        # Just url is equivalent to
        # `brew tap kptdev/kpt https://github.com/kptdev/kpt`
        brew.tap(
            url="https://github.com/kptdev/kpt",
        )

        # src and url is equivalent to
        # `brew tap example/project https://github.example.com/project`
        brew.tap(
            src="example/project",
            url="https://github.example.com/project",
        )

        # Multiple taps
        for tap in ["includeos/includeos", "ktr0731/evans"]:
            brew.tap(
                name={f"Add brew tap {tap}"},
                src=tap,
            )

    """

    if not (src or url):
        host.noop("no tap was specified")
        return

    src = src or str(urllib.parse.urlparse(url).path).strip("/")

    if len(src.split("/")) != 2:
        host.noop("src '{0}' doesn't have two components.".format(src))
        return

    taps = host.get_fact(BrewTaps)
    already_tapped = src in taps

    if present and already_tapped:
        host.noop("tap {0} already exists".format(src))
        return

    if already_tapped:
        yield "brew untap {0}".format(src)
        return

    if not present:
        host.noop("tap {0} does not exist".format(src))
        return

    cmd = "brew tap {0}".format(src)

    if url is not None:
        cmd = " ".join([cmd, url])

    yield cmd
    return
